import logging
from django.core.exceptions import ValidationError
from avocado.query import operators
from avocado.query.validators import Validator, FieldValidator
from avocado.models import DataContext

__all__ = ('BranchParser', 'ConditionParser', 'CompositeParser', 'TreeParser')


BRANCH_TYPES = ('and', 'or')

log = logging.getLogger(__name__)


def has_keys(data, keys):
    "Return true is all keys are present in data."
    for key in keys:
        if key not in data:
            return False
    return True


def is_branch(data):
    "Validates required structure for a branch node."
    return has_keys(data, ('type', 'children'))


def is_condition(data):
    "Validates required structure for a condition node."
    return has_keys(data, keys=('operator', 'value'))


def is_composite(data):
    "Validates required structure for a composite node."
    return 'composite' in data


def get_parser(data):
    if not data or not isinstance(data, dict):
        return
    if is_composite(data):
        return CompositeParser
    elif is_condition(data):
        return ConditionParser
    elif is_branch(data):
        return BranchParser


class BranchParser(Validator):
    "Parser and validator for context branch nodes."

    error_messages = {
        'invalid_branch_type': 'the branch type is invalid',
    }

    warning_messages = {
        'empty_branch': 'the branch contains no children',
        'invalid_child': 'one or more child nodes are invalid',
    }

    error_messages.update(Validator.error_messages)
    warning_messages.update(Validator.warning_messages)

    fields = ('type', 'children')

    def validate_type(self):
        "Validates the branch type."
        branch_type = self.data.get('type', '').lower()
        if branch_type not in BRANCH_TYPES:
            self.error('invalid_branch_type')
        return branch_type

    def validate_children(self):
        """Recurses validation to children in branch.

        Note this occurs regardless if this node is disabled (or if there
        are warnings), to enable downstream handling.
        """
        children = self.data.get('children')
        cleaned_children = []

        if children:
            for child in children:
                parser = get_parser(child)
                if not parser:
                    self.warn('invalid_child')
                    continue
                cleaned_children.append(parser(child, **self.context))
        else:
            self.warn('empty_branch')
        return cleaned_children


class ConditionParser(FieldValidator):
    error_messages = {
        'invalid_value_type': 'invalid value type for field',
        'invalid_operator': 'invalid operator for field',
        'invalid_operator_for_value': 'invalid operator for value',
    }

    warning_messages = {
        'value_out_of_range': 'value out of range',
        'value_not_a_choice': 'value not a choice',
        'field_not_nullable': 'the field is not nullable',
    }

    error_messages.update(FieldValidator.error_messages)
    warning_messages.update(FieldValidator.warning_messages)

    fields = ('concept', 'field', 'value', 'operator', 'nulls')

    def validate_value(self):
        "Checks the value is valid for the field."
        field = self.cleaned_data.get('field')

        if not field:
            return

        try:
            return field.to_python(self.data.get('value'))
        except ValidationError:
            self.error('invalid_value_type')

    def validate_operator(self):
        "Checks the operator is valid for the field"
        field = self.cleaned_data.get('field')

        if not field:
            return

        operator = self.data.get('operator')

        # Check if this is a valid operator for the field
        if operator not in field.operators:
            self.error('invalid_operator')

        # Double check this is also registered (in case the above
        # operators are not built-ins)
        if operator not in operators.registry:
            log.error('Operator {0} locally defined, but not in the registry'
                      .format(operator))
            self.error('invalid_operator')

        return operators.registry.get(operator)

    def validate_nulls(self):
        field = self.cleaned_data.get('field')

        if not field:
            return

        nulls = self.data.get('nulls')
        if nulls and not field.nullable:
            self.warn('field_not_nullable')
        return nulls


class CompositeParser(Validator):
    error_messages = {
        'context_does_not_exist': 'the context does not exist',
    }

    warning_messages = {
        'context_not_defined': 'the context is not defined',
    }

    error_messages.update(Validator.error_messages)
    warning_messages.update(Validator.warning_messages)

    fields = ('context',)

    def validate_context(self):
        context = self.data.get('context')

        if not context:
            self.warn('context_not_defined')

        kwargs = {'pk': context}
        if 'user' in self.context:
            kwargs = {'user': self.context['user']}

        try:
            return DataContext.objects.get(**kwargs)
        except DataContext.DoesNotExist:
            self.error('context_does_not_exist')


class TreeParser(Validator):
    error_messages = {
        'invalid': 'invalid data',
    }

    def validate(self):
        parser = get_parser(self.data)
        if parser:
            self.cleaned_data['tree'] = parser(self.data, **self.context)
        else:
            self.error('invalid')



import inspect
from django.db import models, transaction
from django.db.models.query import QuerySet
from django.contrib.contenttypes.models import ContentType
from avocado.conf import settings
from .utils import get_object_data


class RevisionManager(models.Manager):
    def get_for_model(self, model):
        """Returns all revisions across instances for a model.

        A model is required or a QuerySet can be supplied instead to restrict
        which object revisions are returned.
        """
        if inspect.isclass(model) and issubclass(model, models.Model):
            kwargs = {
                'content_type': ContentType.objects.get_for_model(model),
            }
        elif isinstance(model, QuerySet):
            queryset = model
            model = queryset.model

            kwargs = {
                'object_id__in':
                queryset.values_list('pk', flat=True).order_by(),
                'content_type': ContentType.objects.get_for_model(model),
            }
        else:
            raise ValueError('A model class or QuerySet is required')
        return self.filter(**kwargs)

    def get_for_object(self, obj, model=None):
        "Returns all revisions for the object."
        queryset = None

        if model is not None:
            queryset = self.get_for_model(model)

        if isinstance(obj, models.Model):
            if queryset is None:
                queryset = self.get_for_model(obj.__class__)
            return queryset.filter(object_id=obj.pk)

        if queryset is None:
            raise ValueError('A model or queryset must be supplied if a '
                             'primary key is used')

        return queryset.filter(object_id=obj.pk)

    def latest_for_model(self, model):
        "Returns the latest revision across instances for a model."
        try:
            return self.get_for_model(model).latest()
        except self.model.DoesNotExist:
            pass

    def latest_for_object(self, obj, model=None):
        "Returns the latest revision for the object."
        try:
            return self.get_for_object(obj, model=model).latest()
        except self.model.DoesNotExist:
            pass

    def object_has_changed(self, obj, model=None, fields=None):
        "Returns true if the object data has changed since the last revision."
        revision = self.latest_for_object(obj, model=model)
        if revision is None:
            return True
        return bool(revision.diff(obj, fields=fields))

    @transaction.commit_on_success
    def create_revision(self, obj, fields, model=None, commit=True,
                        deleted=False, **kwargs):
        """Creates a revision for the object.

        The previous revision will be compared against to prevent having
        a redundant revision.
        """
        changed = False
        changes = None

        revision = self.latest_for_object(obj, model=model)

        if not revision or deleted:
            changed = True
        else:
            changes = revision.diff(obj, fields=fields)
            if changes:
                changed = True

        if changed:
            kwargs.setdefault('user', getattr(obj, 'user'))
            kwargs.setdefault('session_key', getattr(obj, 'session_key'))

            if not deleted:
                data = get_object_data(obj, fields)

                if not changes:
                    changes = {}

                    # We purposefully exclude the old_value key here. Since
                    # this is the first revision, there is no old_value so we
                    # will leave it to the clients to check for old_value keys
                    # before referencing it. A missing old_value key is the
                    # indication that this is the first revision.
                    for key in fields:
                        changes[key] = {
                            'new_value': data.get(key),
                        }
            else:
                data = None

            revision = self.model(content_object=obj, data=data,
                                  deleted=deleted, changes=changes, **kwargs)

            if commit:
                revision.save()
            return revision

    @transaction.commit_on_success
    def revert(self, obj, model=None, commit=True):
        "Revert object to last revision."
        revision = self.get_latest_for_object(obj, model=model)
        if revision:
            revision.apply(obj, commit=commit)
        return obj

    @transaction.commit_on_success
    def cull_for_object(self, obj, model=None, max_size=None):
        """Culls revisions for an object down to `max_size`.

        `max_size` defaults to the Avocado setting `HISTORY_MAX_SIZE`.
        """
        if max_size is None:
            max_size = settings.HISTORY_MAX_SIZE
        revisions = self.get_for_object(obj, model=model)\
            .order_by('timestamp').values_list('pk', flat=True)
        self.filter(pk__in=revisions[max_size:]).delete()



from optparse import make_option
from django.core.management.base import BaseCommand
from django.db import transaction
from avocado.conf import settings
from avocado.history.models import Revision


class Command(BaseCommand):
    help = 'Utilities for managing the history API'

    option_list = BaseCommand.option_list + (
        make_option('--cull', action='store_true', help='Culls entries that '
                    'exceeds the maximum allowed history size for each '
                    'instance.'),

        make_option('--max-size', type=int, help='Specify the max history '
                    'size for commands that require it. Defaults to '
                    'HISTORY_MAX_SIZE.'),
    )

    @transaction.commit_on_success
    def handle(self, *args, **options):
        cull = options.get('cull')
        max_size = options.get('max_size', settings.HISTORY_MAX_SIZE)

        if not cull:
            print('No option specified, nothing to do')
            return

        if not max_size:
            print('Unlimited history is allowed, nothing to do.')
            return

        # Get all distinct objects. Empty call to `order_by` removes the
        # ordering to save some overhead.
        objects = Revision.objects.values_list('object_id', 'content_type')\
            .order_by().distinct()

        # Prune history of each object
        for lookup in objects:
            Revision.objects.cull_for_object(*lookup, max_size=max_size)



from copy import deepcopy
from django.db import models


def _validate_field(f):
    "Check if the field is supported for versioning."
    if f.primary_key:
        raise TypeError('Primary keys cannot be versioned.')
    if isinstance(f, (models.ManyToManyField, models.ForeignKey)):
        raise TypeError('Relational fields cannot be versioned.')
    if not f.editable:
        raise TypeError('Non-editable cannot be versioned.')
    return f


def get_object_data(instance, fields):
    "Returns a dict of field data."
    data = {}
    for name in fields:
        _validate_field(instance._meta.get_field(name))
        data[name] = deepcopy(getattr(instance, name))
    return data


def get_model_fields(model):
    "Returns a list of valid fields for versioning."
    fields = []
    for f in model._meta.local_fields:
        # No primary key
        if f is model._meta.pk:
            continue
        # No non-editable fields
        if not f.editable:
            continue
        # No relational fields
        if isinstance(f, (models.ForeignKey, models.ManyToManyField)):
            continue
        fields.append(f.name)
    return fields


def validate_fields(model, fields, exclude):
    "Validates all fields exist and are supported for versioning."
    fields = fields or []
    exclude = exclude or []

    for name in list(fields) + list(exclude):
        _validate_field(model._meta.get_field(name))



import inspect
from functools import wraps
from avocado.conf import settings
from .proxy import CacheProxy

NEVER_EXPIRE = 60 * 60 * 24 * 30  # 30 days
CACHE_KEY_FUNC = lambda l: ':'.join([str(x) for x in l])


def instance_cache_key(instance, label=None, version=None):
    """Creates a cache key for the instance with an optional label and version.
    The instance is uniquely defined based on the app, model and primary key of
    the instance.

    A `label` is used to differentiate cache for an instance.

    The `version` can be a scalar (i.e. a string or int), a function, instance
    property or method.
    """
    if version is None:
        version = '-'
    elif callable(version):
        version = version(instance, label=label)
    elif hasattr(instance, version):
        version = getattr(instance, version)
        if callable(version):
            version = version()

    opts = instance._meta
    key = [opts.app_label, opts.module_name, instance.pk, version]

    if label is not None:
        key.append(label)

    return CACHE_KEY_FUNC(key)


def cached_method(func=None, version=None, timeout=NEVER_EXPIRE,
                  key_func=instance_cache_key):
    "Wraps a method and caches the output indefinitely."

    def decorator(func):
        # Single cache proxy shared across all instances. All methods require
        # the instance to be passed.
        cache_proxy = CacheProxy(func, version, timeout, key_func)

        @wraps(func)
        def inner(self, *args, **kwargs):
            # This check is here to be ensure transparency of the augmented
            # methods below. The agumented methods will be a no-op since the
            # `func_self` will never be set as long as this condition is true.
            if not settings.DATA_CACHE_ENABLED or args or kwargs:
                return func(self, *args, **kwargs)
            return cache_proxy.get_or_set(self, *args, **kwargs)

        # Augment method with a few methods. These are wrapped in a lambda
        # to prevent mucking the cache_proxy instance directly..
        inner.flush = lambda i: cache_proxy.flush(i)
        inner.cached = lambda i: cache_proxy.cached(i)
        inner.cache_key = lambda i: cache_proxy.cache_key(i)

        return inner

    if inspect.isfunction(func):
        return decorator(func)
    return decorator



import logging
from django.core.cache import cache

logger = logging.getLogger(__name__)


class CacheProxy(object):
    def __init__(self, func, version, timeout, key_func):
        self.func = func
        self.label = func.__name__
        self.version = version
        self.timeout = timeout
        self.key_func = key_func

    def cache_key(self, instance):
        return self.key_func(instance, label=self.label, version=self.version)

    def _set(self, key, data):
        logger.debug('Compute property cache "{0}"'.format(key))
        if data is not None:
            cache.set(key, data, timeout=self.timeout)
            logger.debug('Set property cache "{0}"'.format(key))

    def get(self, instance):
        key = self.cache_key(instance)
        data = cache.get(key)
        logger.debug('Get property cache "{0}"'.format(key))
        return data

    def get_or_set(self, instance, *args, **kwargs):
        # Reference to prevent the key from being changed mid-execution
        key = self.cache_key(instance)

        data = cache.get(key)
        if data is None:
            data = self.func(instance, *args, **kwargs)
            self._set(key, data)
        return data

    def flush(self, instance):
        "Flushes cached data for this method."
        key = self.cache_key(instance)
        cache.delete(key)
        logger.debug('Delete property cache "{0}"'.format(key))

    def cached(self, instance):
        "Checks if the data is in the cache."
        return self.cache_key(instance) in cache



from random import choice
from string import ascii_lowercase, digits
from django import forms
from django.contrib.auth.models import User
from django.db import models
from django.utils.importlib import import_module
from avocado.conf import settings


# 41 characters @ 30 characters per username = 3.16 billion permutations
# I think that will cover it..
USERNAME_CHARS = ascii_lowercase + digits + '@.+-_'


def get_form_class(name):
    # Absolute import if a period exists, otherwise assume the
    # name refers to a built-in Django class
    if '.' in name:
        path = name.split('.')[:-1]
        module = import_module(path)
    else:
        if not name.endswith('Field'):
            name = name + 'Field'
        module = forms
    return getattr(module, name)


def get_internal_type(field):
    "Get model field internal type with 'field' off."
    datatype = field.get_internal_type().lower()
    if datatype.endswith('field'):
        datatype = datatype[:-5]
    return datatype


def get_simple_type(internal):
    """Returns a simple type mapped from the internal type."

    By default, it will use the field's internal type, but can be
    overridden by the ``SIMPLE_TYPES`` setting.
    """
    if isinstance(internal, models.Field):
        internal = get_internal_type(internal)
    return settings.SIMPLE_TYPES.get(internal, internal)


def get_heuristic_flags(field):
    # TODO add better conditions for determining how to set the
    # flags for most appropriate interface.
    # - Determine length of MAX value for string-based fields to rather
    # than relying on the `max_length`. This will enable checking TextFields
    # - Numerical fields may be enumerable, check the size of them if an
    # option is set?
    # For strings and booleans, set the enumerable flag by default
    # it below the enumerable threshold
    # TextFields are typically used for free text
    enumerable = False

    if field.internal_type != 'text' \
            and field.simple_type in ('string', 'boolean') \
            and field.size() <= settings.ENUMERABLE_MAXIMUM:
        enumerable = True

    return {
        'enumerable': enumerable,
    }


def parse_field_key(key):
    "Returns a field lookup based on a variety of key types."
    if isinstance(key, int):
        return {'pk': key}
    keys = ('app_name', 'model_name', 'field_name')
    if isinstance(key, models.Field):
        opts = key.model._meta
        toks = [opts.app_label, opts.module_name, key.name]
    elif isinstance(key, basestring):
        toks = key.split('.')
    elif isinstance(key, (list, tuple)):
        toks = key
    offset = len(keys) - len(toks)
    return dict(zip(keys[offset:], toks))


def generate_random_username(length=30, max_attempts=100):
    for i in xrange(max_attempts):
        username = ''.join(choice(USERNAME_CHARS) for i in xrange(length))
        if not User.objects.filter(username=username).exists():
            return username
    raise ValueError('Maximum attempts made to generate username')


def create_email_based_user(email):
    """
    Creates an inactive user from the email address. These users are
    placeholders for those users that do not have accounts. This is initially
    planned for use in conjunction with adding users to DataQuery.shared_users.
    """
    username = generate_random_username()
    email = User.objects.normalize_email(email)

    user = User(username=username, email=email, is_active=False)
    user.set_unusable_password()
    user.full_clean()
    user.save()

    return user



from warnings import warn
from modeltree.tree import trees
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from avocado.core import utils

AND = 'AND'
OR = 'OR'
BRANCH_KEYS = ('children', 'type')
CONDITION_KEYS = ('operator', 'value')
COMPOSITE_KEYS = ('composite',)
LOGICAL_OPERATORS = ('and', 'or')


def has_keys(obj, keys):
    "Check the required keys are present in `obj`"
    for key in keys:
        if key not in obj:
            return False
    return True


def is_branch(obj):
    "Validates required structure for a branch node"
    if has_keys(obj, keys=BRANCH_KEYS):
        return True


def is_condition(obj):
    "Validates required structure for a condition node"
    if has_keys(obj, keys=CONDITION_KEYS):
        if 'field' in obj or 'id' in obj:
            return True


def is_composite(obj):
    if has_keys(obj, keys=COMPOSITE_KEYS):
        return True


class Node(object):
    condition = None
    annotations = None
    extra = None
    language = None

    def __init__(self, tree=None, **context):
        self.tree = tree
        self.context = context

    def apply(self, queryset=None, distinct=True):
        if queryset is None:
            queryset = trees[self.tree].get_queryset()
        if self.annotations:
            queryset = queryset.values('pk').annotate(**self.annotations)
        if self.condition:
            queryset = queryset.filter(self.condition)
        if self.extra:
            queryset = queryset.extra(**self.extra)
        if distinct:
            queryset = queryset.distinct()
        return queryset


class Condition(Node):
    "Contains information for a single query condition."
    def __init__(self, value, operator, id=None, field=None,
                 concept=None, **context):

        if field:
            self.field_key = field
        else:
            self.field_key = id
            warn('The "id" key has been replaced with "field"',
                 DeprecationWarning)

        self.concept_key = concept
        self.operator = operator
        self.value = value

        super(Condition, self).__init__(**context)

    @property
    def _meta(self):
        if not hasattr(self, '__meta'):
            self.__meta = self.field.translate(operator=self.operator,
                                               value=self.value,
                                               tree=self.tree, **self.context)
        return self.__meta

    @property
    def concept(self):
        if not hasattr(self, '_concept'):
            if self.concept_key:
                from avocado.models import DataConcept
                self._concept = DataConcept.objects.get(id=self.concept_key)
            else:
                self._concept = None
        return self._concept

    @property
    def field(self):
        if not hasattr(self, '_field'):
            from avocado.models import DataField
            # Parse to get into a consistent format
            field_key = utils.parse_field_key(self.field_key)

            if self.concept:
                self._field = self.concept.fields.get(**field_key)
            else:
                self._field = DataField.objects.get(**field_key)
        return self._field

    @property
    def condition(self):
        return self._meta['query_modifiers'].get('condition', None)

    @property
    def annotations(self):
        return self._meta['query_modifiers'].get('annotations', None)

    @property
    def extra(self):
        return self._meta['query_modifiers'].get('extra', None)

    @property
    def language(self):
        meta = self._meta.copy()
        meta.pop('query_modifiers')
        cleaned = meta.pop('cleaned_data')
        meta['language'] = cleaned['language']
        return meta


class Branch(Node):
    "Provides a logical relationship between it's children."
    def __init__(self, type, **context):
        self.type = (type.upper() == AND) and AND or OR
        self.children = []
        super(Branch, self).__init__(**context)

    def _combine(self, q1, q2):
        if self.type.upper() == OR:
            return q1 | q2
        return q1 & q2

    @property
    def condition(self):
        if not hasattr(self, '_condition'):
            condition = None
            for node in self.children:
                if node.condition:
                    if condition:
                        condition = self._combine(node.condition, condition)
                    else:
                        condition = node.condition
            self._condition = condition
        return self._condition

    @property
    def annotations(self):
        if not hasattr(self, '_annotations'):
            self._annotations = {}
            for node in self.children:
                if node.annotations:
                    self._annotations.update(node.annotations)
        return self._annotations

    @property
    def extra(self):
        extra = {}
        for node in self.children:
            if not node.extra:
                continue
            for key, value in node.extra.items():
                _type = type(value)
                # Initialize an empty container for the value type..
                extra.setdefault(key, _type())
                if _type is list:
                    current = extra[key][:]
                    [extra[key].append(x) for x in value if x not in current]
                elif _type is dict:
                    extra[key].update(value)
                else:
                    raise TypeError('The ".extra()" method only takes list of '
                                    'dicts as keyword values')
        return extra

    @property
    def language(self):
        out = {'type': self.type.lower(), 'children': []}
        for node in self.children:
            out['children'].append(node.language)
        return out


def validate(attrs, **context):
    if not attrs:
        return None

    if type(attrs) is not dict:
        raise ValidationError('Object must be of type dict')

    enabled = attrs.pop('enabled', None)

    attrs.pop('errors', None)
    attrs.pop('warnings', None)
    errors = []
    warnings = []

    if is_composite(attrs):
        from avocado.models import DataContext
        try:
            if 'user' in context:
                cxt = DataContext.objects.get(id=attrs['composite'],
                                              user=context['user'])
            else:
                cxt = DataContext.objects.get(id=attrs['composite'])
            validate(cxt.json, **context)
            attrs['language'] = cxt.name
        except DataContext.DoesNotExist:
            enabled = False
            errors.append(u'DataContext "{0}" does not exist.'
                          .format(attrs['id']))

    elif is_condition(attrs):
        from avocado.models import DataField, DataConcept
        field_key = attrs.get('field', attrs.get('id'))
        # Parse to get into a consistent format
        field_key = utils.parse_field_key(field_key)

        try:
            if 'concept' in attrs:
                concept = DataConcept.objects.get(id=attrs['concept'])
                field = concept.fields.get(**field_key)
            else:
                field = DataField.objects.get(**field_key)
            field.validate(operator=attrs['operator'], value=attrs['value'])
            node = parse(attrs, **context)
            attrs['language'] = node.language['language']
        except ObjectDoesNotExist:
            enabled = False
            errors.append('Field does not exist')

    elif is_branch(attrs):
        if attrs['type'] not in LOGICAL_OPERATORS:
            enabled = False
        else:
            map(lambda x: validate(x, **context), attrs['children'])
    else:
        enabled = False

    # If this condition was originally disabled, ensure that decision is
    # persisted
    if enabled is False:
        attrs['enabled'] = False

    if errors:
        attrs['errors'] = errors
    if warnings:
        attrs['warnings'] = warnings

    return attrs


def parse(attrs, **context):
    if not attrs or attrs.get('enabled') is False:
        node = Node(**context)
    elif is_composite(attrs):
        from avocado.models import DataContext
        if 'user' in context:
            cxt = DataContext.objects.get(id=attrs['composite'],
                                          user=context['user'])
        else:
            cxt = DataContext.objects.get(id=attrs['composite'])
        return parse(cxt.json, **context)
    elif is_condition(attrs):
        node = Condition(operator=attrs['operator'], value=attrs['value'],
                         id=attrs.get('id'), field=attrs.get('field'),
                         **context)
    else:
        node = Branch(type=attrs['type'], **context)
        node.children = map(lambda x: parse(x, **context), attrs['children'])
    return node



from warnings import warn
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from modeltree.tree import trees
from modeltree.query import ModelTreeQuerySet


SORT_DIRECTIONS = ('asc', 'desc')


class Node(object):
    def __init__(self, facets=None, **context):
        self.facets = facets or []
        self.tree = context.pop('tree', None)
        self.context = context

    @property
    def concept_ids(self):
        ids = []
        for facet in self.facets:
            if facet.get('enabled') is False:
                continue
            if facet.get('visible') is False:
                continue
            ids.append(facet['concept'])
        return ids

    @property
    def ordering(self):
        ids = []
        length = len(self.facets)
        for facet in self.facets:
            if facet.get('enabled') is False:
                continue
            if facet.get('sort') in SORT_DIRECTIONS:
                # If sort is not defined, default to length of facets
                ids.append((
                    facet.get('sort_index', length),
                    facet['concept'],
                    facet['sort'],
                ))
        # Sort relative to sort index
        ids.sort(key=lambda x: x[0])
        # Return only the concept id and sort direction
        return [(c, s) for i, c, s in ids]

    def _get_concepts(self, ids):
        "Returns an ordered list of concepts based on `ids`."
        if not ids:
            return []

        from avocado.models import DataConcept

        concepts = list(DataConcept.objects.filter(pk__in=ids))
        concepts.sort(key=lambda o: ids.index(o.pk))
        return concepts

    def _get_fields_for_concepts(self, ids):
        "Returns an ordered list of fields for concept `ids`."
        from avocado.models import DataConceptField
        if not ids:
            return OrderedDict()
        # Concept fields that are sorted by concept then order, but are not
        # in the original order defined in `ids`
        cfields = list(DataConceptField.objects.filter(concept__pk__in=ids)
                       .select_related()
                       .order_by('concept', 'order'))

        # Order concept fields relative to `ids`
        cfields.sort(key=lambda o: ids.index(o.concept.pk))

        # Construct an ordered dict of fields by their concept
        groups = OrderedDict()
        for cf in cfields:
            pk = cf.concept.pk
            if pk not in groups:
                groups[pk] = []
            groups[pk].append(cf.field)
        return groups

    def _get_select(self, distinct):
        # Apply all fields to the query to ensure ordering get applied.
        # Django removes ORDER BY statements if column is not present in
        # SELECT since it will cause a SQL error. This ensures the ordering
        # is applied at the SQL level. The caveat here is that the rows
        # returned will include this extra data. The exporter classes handle
        # this by removing redundant rows relative to the *original* columns.
        ids = list(self.concept_ids)
        ordering = self.ordering
        if ordering and distinct:
            ids += list(zip(*ordering)[0])

        # Flatten the grouped fields
        fields = [i for l in self._get_fields_for_concepts(ids).values()
                  for i in l]

        model_fields = []

        for f in fields:
            if f.lexicon:
                model_fields.append(f.model._meta.get_field('label'))
            else:
                model_fields.append(f.field)
        return model_fields

    def _get_order_by(self):
        "Returns directional lookups to be unpacked in `QuerySet.order_by`."
        order_by = []
        ordering = self.ordering

        if ordering:
            tree = trees[self.tree]
            ids, directions = zip(*ordering)
            groups = self._get_fields_for_concepts(ids)

            for pk, direction in ordering:
                for f in groups[pk]:
                    # Special case for Lexicon-based models, order by their
                    # model-defined `order` field.
                    if f.lexicon:
                        field = f.model._meta.get_field('order')
                        lookup = tree.query_string_for_field(field)
                    else:
                        lookup = tree.query_string_for_field(f.field)

                    if direction.lower() == 'desc':
                        order_by.append('-' + lookup)
                    else:
                        order_by.append(lookup)

        return order_by

    # Primary method for apply this view to a QuerySet
    def apply(self, queryset=None, include_pk=True):
        tree = trees[self.tree]
        if queryset is None:
            queryset = tree.get_queryset()

        # Convert to a ModelTreeQuerySet for the `select()` method
        queryset = ModelTreeQuerySet(tree, query=queryset.query)

        # Set model fields for `select()` method
        select = self._get_select(queryset.query.distinct)
        queryset = queryset.select(*select, include_pk=include_pk)

        # Set the order by on the QuerySet
        order_by = self._get_order_by()
        if order_by:
            queryset = queryset.order_by(*order_by)

        return queryset

    # Additional public methods for general use and interrogation
    def get_concepts_for_select(self):
        return self._get_concepts(self.concept_ids)

    def get_fields_for_select(self):
        return self._get_fields_for_concepts(self.concept_ids)

    def get_concepts_for_order_by(self):
        ids = []
        ordering = self.ordering
        if ordering:
            ids = zip(*ordering)[0]
        return self._get_concepts(ids)

    def get_fields_for_order_by(self):
        ids = []
        ordering = self.ordering
        if ordering:
            ids = zip(*ordering)[0]
        return self._get_fields_for_concepts(ids)


def convert_legacy(attrs):
    facets = []
    columns = attrs.get('columns', [])
    ordering = attrs.get('ordering', [])

    # Map to position
    _ordering = {}
    for i, (concept, direction) in enumerate(ordering):
        _ordering[concept] = i

    for pk in columns:
        facet = {'concept': pk}

        if pk in _ordering:
            index = _ordering.pop(pk)
            facet['sort'] = ordering[index][1]
            facet['sort_index'] = index

        facets.append(facet)

    # Append the sort only concepts
    for pk in _ordering:
        index = _ordering[pk]
        facets.append({
            'visible': False,
            'concept': pk,
            'sort': ordering[index][1],
            'sort_index': index,
        })

    return facets


def validate(facets, **context):
    if not facets:
        return None

    from avocado.models import DataConcept

    # Legacy format
    if isinstance(facets, dict):
        warn('The dict-based view structure has been deprecated. '
             'A list of facets objects must now be provided.',
             DeprecationWarning)
        facets = convert_legacy(facets)

    for attrs in facets:
        enabled = attrs.pop('enabled', None)
        attrs.pop('errors', None)
        attrs.pop('warnings', None)
        errors = []
        warnings = []

        concept = None

        if 'concept' not in attrs:
            enabled = False
            errors.append('Concept is required')
        else:
            try:
                concept = DataConcept.objects.get(pk=attrs.get('concept'))
            except DataConcept.DoesNotExist:
                enabled = False
                errors.append('Concept does not exist')

        if 'sort' in attrs and attrs['sort'] not in SORT_DIRECTIONS:
            warnings.append('Invalid sort direction. Must be "asc" or "desc"')

        if concept and not concept.sortable:
            warnings.append('Cannot sort by concept')

        if enabled is False:
            attrs['enabled'] = False

        # Amend errors and warnings if present
        if errors:
            attrs['errors'] = errors

        if warnings:
            attrs['warnings'] = warnings

    return facets


def parse(facets, **context):
    if not facets:
        return Node(**context)

    # Legacy format
    if isinstance(facets, dict):
        warn('The dict-based view structure has been deprecated. '
             'A list of facets objects must now be provided.',
             DeprecationWarning)
        facets = convert_legacy(facets)

    return Node(facets, **context)



import jsonfield
from datetime import datetime
from django.db import models
from django.contrib.contenttypes import generic
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType


class Log(models.Model):
    # Reference to an object if one is applicable
    content_type = models.ForeignKey(ContentType, null=True, blank=True)
    object_id = models.PositiveIntegerField(null=True, blank=True)
    content_object = generic.GenericForeignKey()

    # The event name. This must be unique for the event type to ensure
    # data can be aggregated for analysis
    event = models.CharField(max_length=200)

    # Extra data for the particular log entry. This must be JSON serializable.
    data = jsonfield.JSONField(null=True, blank=True)

    # The user and/or session that caused the event
    user = models.ForeignKey(User, null=True, blank=True, related_name='+')
    session_key = models.CharField(max_length=40, null=True, blank=True)

    # The timestamp of the event
    timestamp = models.DateTimeField(default=datetime.now)

    class Meta(object):
        app_label = 'avocado'

    def __unicode__(self):
        if self.content_type:
            label = '{0}:{1}'.format(self.content_type.model, self.event)
            if self.object_id:
                label = '{0} for {1}'.format(label, self.content_object)
            return label
        return self.event



from .model import instance_cache_key, cached_method  # noqa
from .receivers import post_save_cache, pre_delete_uncache  # noqa
from .managers import CacheManager  # noqa
from .query import CacheQuerySet  # noqa
from .proxy import CacheProxy  # noqa



from avocado.models import DataConcept, DataView
from avocado.formatters import Formatter
from cStringIO import StringIO


class BaseExporter(object):
    "Base class for all exporters"
    file_extension = 'txt'
    content_type = 'text/plain'
    preferred_formats = []

    def __init__(self, concepts=None):
        if concepts is None:
            concepts = ()
        elif isinstance(concepts, DataView):
            node = concepts.parse()
            concepts = node.get_concepts_for_select()

        self.params = []
        self.row_length = 0
        self.concepts = concepts

        for concept in concepts:
            self.add_formatter(concept)

    def __repr__(self):
        return u'<{0}: {1}/{2}>'.format(self.__class__.__name__,
                                        len(self.params), self.row_length)

    def add_formatter(self, formatter, length=None, index=None):
        if isinstance(formatter, DataConcept):
            length = formatter.concept_fields.count()
            formatter = formatter.format
        elif isinstance(formatter, Formatter):
            length = len(formatter.keys)
        elif length is None:
            raise ValueError('A length must be supplied with the to '
                             'denote how much of the row will be formatted.')

        params = (formatter, length)
        self.row_length += length

        if index is not None:
            self.params.insert(index, params)
        else:
            self.params.append(params)

    def get_file_obj(self, name=None):
        if name is None:
            return StringIO()
        if isinstance(name, basestring):
            return open(name, 'w+')
        return name

    def _format_row(self, row, **kwargs):
        for formatter, length in self.params:
            values, row = row[:length], row[length:]
            yield formatter(values, preferred_formats=self.preferred_formats,
                            **kwargs)

    def read(self, iterable, force_distinct=True, offset=None, limit=None,
             *args, **kwargs):
        """Takes an iterable that produces rows to be formatted.

        If `force_distinct` is true, rows will be filtered based on the slice
        of the row that is going to be formatted.

        If `offset` is defined, only rows that are produced after the offset
        index are returned.

        If `limit` is defined, once the limit has been reached (or the
        iterator is exhausted), the loop will exit.
        """
        emitted = 0
        unique_rows = set()

        for i, row in enumerate(iterable):
            if limit is not None and emitted >= limit:
                break

            _row = row[:self.row_length]

            if force_distinct:
                _row_hash = hash(tuple(_row))

                if _row_hash in unique_rows:
                    continue

                unique_rows.add(_row_hash)

            if offset is None or i >= offset:
                emitted += 1
                yield self._format_row(_row, **kwargs)

    def write(self, iterable, *args, **kwargs):
        for row_gen in self.read(iterable, *args, **kwargs):
            row = []
            for data in row_gen:
                row.extend(data.values())
            yield tuple(row)



import logging
from datetime import datetime
from threading import Thread
from avocado.conf import settings
from django.contrib.contenttypes.models import ContentType
from .models import Log

logger = logging.getLogger(__name__)


# This is intended to be strictly for testing purposes
FORCE_SYNC_LOG = getattr(settings, 'FORCE_SYNC_LOG', False)


def _log(instance=None, model=None, **kwargs):
    if instance:
        kwargs['content_object'] = instance
    elif model:
        kwargs['content_type'] = ContentType.objects.get_for_model(model)

    try:
        message = Log(**kwargs)
        message.save()
    except Exception:
        logger.exception('Error logging usage')


def log(event, async=True, **kwargs):
    """Log an event with an optional associated object.

    If `instance` is present, the generic `content_object` will be set. Other
    if just the model is provided, the content type will be associated.

    If a `request` object is supplied, the `user` and/or `session_key` will
    be extracted from it.

    By default the call will be executed in a thread to not block the current
    executing thread.
    """
    kwargs['event'] = event

    if 'request' in kwargs:
        request = kwargs.pop('request')
        if hasattr(request, 'user') and request.user.is_authenticated():
            kwargs['user'] = request.user
        if hasattr(request, 'session'):
            kwargs['session_key'] = request.session.session_key

    # Set the timestamp now to prevent delay due to a blocked thread
    if 'timestamp' not in kwargs:
        kwargs['timestamp'] = datetime.now()

    if async and not FORCE_SYNC_LOG:
        thread = Thread(target=_log, kwargs=kwargs)
        thread.start()
    else:
        _log(**kwargs)



from __future__ import print_function
import re
from django.db.models import get_model
from django.core.management.base import BaseCommand, CommandError
from avocado.models import DataField
from avocado.lexicon.managers import coerce_float

non_alnum = re.compile(r'[^a-zA-Z0-9\-]+')

_help = """\
Creates a lexicon derived from `field` into the Lexicon
subclass `model`. The distinct list of values for `field`
will be queried and loaded into the table.
"""


class Command(BaseCommand):
    __doc__ = help = _help

    args = 'field model'

    def handle(self, field_label, model_label, **options):
        model = get_model(*model_label.split('.'))

        if model is None:
            raise CommandError(u'Not model named {0} was found'.format(
                model_label))

        toks = field_label.split('.')
        if len(toks) != 3:
            raise CommandError(u'{0} is not a valid field identifier. Use a '
                               '"." delimited notation, e.g. "app.model.field"'
                               .format(field_label))

        # Not required to be persisted to the database..
        f = DataField(app_name=toks[0], model_name=toks[1], field_name=toks[2])

        if not f.field:
            raise CommandError(u'The field {0} could not be found.'.format(
                field_label))

        count = 0
        values = list(f.values())
        values.sort(key=coerce_float)

        for value in values:
            if value is None or value == '':
                continue
            label = non_alnum.sub(value, ' ').title()
            obj = model(label=label, value=value, order=count, code=count)
            obj.save()
            count += 1

        print(u'{0} distinct values loaded'.format(count))



from django.utils.importlib import import_module
from modeltree.tree import trees
from avocado.formatters import RawFormatter
from avocado.conf import settings

QUERY_PROCESSOR_DEFAULT_ALIAS = 'default'


class QueryProcessor(object):
    """Prepares and builds a QuerySet for export.

    Overriding or extending these methods enable customizing the behavior
    pre/post-construction of the query.
    """
    def __init__(self, context=None, view=None, tree=None, include_pk=True):
        self.context = context
        self.view = view
        self.tree = tree
        self.include_pk = include_pk

    def get_queryset(self, queryset=None, **kwargs):
        "Returns a queryset based on the context and view."
        if self.context:
            queryset = \
                self.context.apply(queryset=queryset, tree=self.tree)

        if self.view:
            queryset = self.view.apply(queryset=queryset, tree=self.tree,
                                       include_pk=self.include_pk)

        if queryset is None:
            queryset = trees[self.tree].get_queryset().values('pk')

        return queryset

    def get_exporter(self, klass, **kwargs):
        "Returns an exporter prepared for the queryset."
        exporter = klass(self.view)

        if self.include_pk:
            pk_name = trees[self.tree].root_model._meta.pk.name
            exporter.add_formatter(RawFormatter(keys=[pk_name]), index=0)

        return exporter

    def get_iterable(self, offset=None, limit=None, **kwargs):
        "Returns an iterable that can be used by an exporter."
        queryset = self.get_queryset(**kwargs)

        if offset is not None and limit is not None:
            queryset = queryset[offset:offset + limit]
        elif offset is not None:
            queryset = queryset[offset:]
        elif limit is not None:
            queryset = queryset[:limit]

        # ModelTreeQuerySet has a raw method defined, but fallback
        # to the creating a results iter if not present.
        if hasattr(queryset, 'raw'):
            iterable = queryset.raw()
        else:
            compiler = queryset.query.get_compiler(queryset.db)
            iterable = compiler.results_iter()

        return iterable


class QueryProcessors(object):
    def __init__(self, processors):
        self.processors = processors
        self._processors = {}

    def __getitem__(self, key):
        return self._get(key)

    def __len__(self):
        return len(self._processors)

    def __nonzero__(self):
        return True

    def _get(self, key):
        # Import class if not cached
        if key not in self._processors:
            toks = self.processors[key].split('.')
            klass_name = toks.pop()
            path = '.'.join(toks)
            klass = getattr(import_module(path), klass_name)
            self._processors[key] = klass
        return self._processors[key]

    @property
    def default(self):
        return self[QUERY_PROCESSOR_DEFAULT_ALIAS]


query_processors = QueryProcessors(settings.QUERY_PROCESSORS)



import jsonfield
from copy import deepcopy
from datetime import datetime
from django.db import models
from django.contrib.contenttypes import generic
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from .managers import RevisionManager
from .utils import get_object_data


class Revision(models.Model):
    # Generic foreign key to the object
    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField(db_index=True)
    content_object = generic.GenericForeignKey()

    # The serialized data for this revision
    data = jsonfield.JSONField(null=True, blank=True)

    # The user that created the revision
    user = models.ForeignKey(User, null=True, blank=True,
                             related_name='+revision')
    session_key = models.CharField(max_length=40, null=True, blank=True)

    # The timestamp of the revision
    timestamp = models.DateTimeField(default=datetime.now, db_index=True)

    # This revision represents a deleted object
    deleted = models.BooleanField(default=False)

    # The changes from the previous revision
    changes = jsonfield.JSONField(null=True, blank=True)

    objects = RevisionManager()

    class Meta(object):
        app_label = 'avocado'
        ordering = ('-timestamp',)
        get_latest_by = 'timestamp'

    def __unicode__(self):
        return '{0} @ {1}'.format(self.content_object, self.timestamp)

    def apply(self, instance, commit=True):
        "Applies this revision to an instance."
        for key, value in self.data.items():
            setattr(instance, key, deepcopy(value))
        if commit:
            instance.save()
        return instance

    def diff(self, instance, fields=None, reverse=False):
        """
        Returns the diff between the revision and the instance, relative
        to the revision. Changed values will include the old and new values
        under the 'old_value' and 'new_value' keys respectively.

        If `reverse` is true, the diff will be relative to the instance.
        If `fields` is not defined, the revision data keys will be used.
        """
        diff = {}

        if not fields:
            fields = self.data.keys()

        data = get_object_data(instance, fields=fields)

        for key in fields:
            if data.get(key) != self.data.get(key):
                if reverse:
                    old_value = data.get(key)
                    new_value = self.data.get(key)
                else:
                    old_value = self.data.get(key)
                    new_value = data.get(key)
                diff[key] = {
                    'old_value': old_value,
                    'new_value': new_value
                }
        return diff



from django.core.management.base import BaseCommand, CommandError
from .utils import get_fields_by_label


class DataFieldCommand(BaseCommand):
    args = '[app | app.model | app.model.field]*'

    def handle_fields(self, *args, **kwargs):
        raise NotImplemented('Subclasses must define this method.')

    def handle(self, *labels, **kwargs):
        try:
            fields = get_fields_by_label(labels)
        except Exception, e:
            raise CommandError(e.message)
        return self.handle_fields(fields, **kwargs)



from django.db.models.signals import pre_delete, post_save
from avocado.core.loader import AlreadyRegistered, NotRegistered
from .receivers import pre_delete_revision, post_save_revision
from . import utils


registry = {}


def register(model, fields=None, exclude=None):
    """Registers a model with the specified fields to be automatically
    versioned on save and delete operations.
    """
    if model in registry:
        raise AlreadyRegistered(u'The model {0} is already registered'.format(
            model.__name__))

    if not fields:
        fields = utils.get_model_fields(model)
    if not exclude:
        exclude = ()

    utils.validate_fields(model, fields, exclude)
    fields = tuple(set(fields) - set(exclude))

    if not fields:
        raise ValueError(u'No fields defined for versioning.')

    dispatch_uid = '{0}_revision'.format(model.__name__)

    pre_delete.connect(pre_delete_revision, weak=False, sender=model,
                       dispatch_uid=dispatch_uid)
    post_save.connect(post_save_revision, weak=False, sender=model,
                      dispatch_uid=dispatch_uid)

    registry[model] = {
        'fields': fields,
        'dispatch_uid': dispatch_uid,
    }


def unregister(model):
    "Unregisters a model from versioning."
    if model not in registry:
        raise NotRegistered(u'The model {0} is not registered'.format(
            model.__name__))
    cache = registry.pop(model)
    pre_delete.disconnect(sender=model, dispatch_uid=cache['dispatch_uid'])
    post_save.disconnect(sender=model, dispatch_uid=cache['dispatch_uid'])



from django.core.cache import cache
from django.db.models.query import QuerySet
from .model import CACHE_KEY_FUNC

PK_LOOKUPS = ('pk', 'pk__exact')


class CacheQuerySet(QuerySet):
    def filter(self, *args, **kwargs):
        """For primary-key-based lookups, instances may be cached to prevent
        excessive database hits. If this is a primary-key lookup, the cache
        will be checked and populated in the `_result_cache` if available.
        """
        clone = super(CacheQuerySet, self).filter(*args, **kwargs)

        pk = None
        opts = self.model._meta
        pk_name = opts.pk.name

        # Look for `pk` and the actual name of the primary key field
        for key in list(PK_LOOKUPS) + [pk_name, u'{0}__exact'.format(pk_name)]:
            if key in kwargs:
                pk = kwargs[key]
                break

        if pk is not None:
            key = CACHE_KEY_FUNC([opts.app_label, opts.module_name, pk])
            obj = cache.get(key)
            if obj is not None:
                clone._result_cache = [obj]

        return clone



from django.core.cache import cache
from .model import instance_cache_key, NEVER_EXPIRE


def post_save_cache(sender, instance, **kwargs):
    """General post-save handler for caching model instances. NOTE: This must
    be used in conjunction with the `pre_delete_uncache` since the cache is set
    to never expire.
    """
    cache.set(instance_cache_key(instance), instance, timeout=NEVER_EXPIRE)


def pre_delete_uncache(sender, instance, **kwargs):
    "General post-delete handler for removing cache for model instances."
    cache.delete(instance_cache_key(instance))



import logging
from django.core.exceptions import ValidationError
from avocado.core import utils
from avocado.conf import settings
from avocado.models import DataConcept, DataField

log = logging.getLogger(__name__)


class Validator(object):
    error_messages = {}
    warning_messages = {}

    fields = ()

    def __init__(self, data, **context):
        # Copy of the input data, this will be updated as necessary reflecting
        # errors, warnings, and other changes.
        self.data = data.copy()

        # This represents the _cleaned_ representations of the input such
        # as model instances. Validators that depend on this upstream data
        # can check this to ensure the data has been cleaned.
        self.cleaned_data = {}

        # Contains the additional context pass into the validator that may be
        # used during validation.
        self.context = context

        self.errors = []
        self.warnings = []

    def _validate(self):
        """
        Performs valiation of fields defined in `fields`. Output will added to
        `cleaned_data`. Calls to `error` will prevent the output from being
        added.
        """
        for field in self.fields:
            method_name = 'validate_{0}'.format(field)
            if hasattr(self, method_name):
                try:
                    result = getattr(self, method_name)()
                    self.cleaned_data[field] = result
                except ValidationError:
                    pass

    def _post_validate(self):
        """Performs post-validation. If errors or warnings are present, the
        `enabled` flag gets toggled to reflect this.
        """
        # Remove previous annotations
        self.data.pop('errors', [])
        enabled = self.data.pop('enabled', True)
        warnings = self.data.pop('warnings', [])

        if self.errors:
            self.data['errors'] = self.errors
            enabled = False

        if self.warnings:
            self.data['warnings'] = self.warnings
            # If the warnings have changed from the previous state, force
            # the state to be disabled.
            if set(self.warnings) != set(warnings):
                enabled = False

        # Only set the flag if false or if warnings are still around for
        # revalidation.
        if not enabled or self.warnings:
            self.data['enabled'] = enabled

        # Copy data, update with cleaned output. This will expose the
        # objects
        cleaned_data = self.cleaned_data
        self.cleaned_data = self.data.copy()
        self.cleaned_data.update(cleaned_data)

    @property
    def validation_errors(self):
        messages = []
        for key in self.errors:
            if key in settings.VALIDATION_ERRORS:
                messages.append(settings.VALIDATION_ERRORS[key])
            elif key in self.validation_errors:
                messages.append(self.validation_errors[key])
            raise KeyError('No message for key: {0}'.format(key))
        return messages

    @property
    def validation_warnings(self):
        messages = []
        for key in self.warnings:
            if key in settings.VALIDATION_WARNINGS:
                messages.append(settings.VALIDATION_WARNINGS[key])
            elif key in self.validation_warnings:
                messages.append(self.validation_warnings[key])
            raise KeyError('No message for key: {0}'.format(key))
        return messages

    def error(self, key):
        "Adds an error message based on a message key."
        if key not in self.error_messages:
            raise KeyError('No validation error with key: {0}'.format(key))
        self.errors.append(key)
        raise ValidationError('noop')

    def warn(self, key):
        "Adds an warning message based on a message key."
        if key not in self.warning_messages:
            raise KeyError('No validation warning with key: {0}'.format(key))
        self.warnings.append(key)

    def validate(self):
        "Performs cross-data cleaning."

    def is_valid(self):
        "Performs full validation."
        self._validate()
        self.validate()
        self._post_validate()
        if self.errors:
            return False
        return True


class FieldValidator(Validator):
    error_messages = {
        'field_required': 'field required',
        'field_does_not_exist': 'the field does not exist',
        'field_does_not_exist_for_concept': 'the field does not exist for the '
                                            'specified concept',
        'ambiguous_field': 'the field lookup is ambiguous',
        'ambiguous_field_for_concept': 'the field lookup for the specified is '
                                       'ambiguous',
        'concept_does_not_exist': 'the concept does not exist',
    }

    error_messages.update(Validator.error_messages)

    fields = ('concept', 'field')

    def validate_concept(self):
        concept = self.data.get('concept')

        if not concept:
            return

        # Use the `ident` field if the concept is defined as a string
        if isinstance(concept, basestring):
            kwargs = {'ident': concept}
        else:
            kwargs = {'pk': concept}

        if 'user' in self.context:
            queryset = DataConcept.objects.published(user=self.context['user'])
        else:
            queryset = DataConcept.objects.all()

        try:
            return queryset.get(**kwargs)
        except DataConcept.DoesNotExist:
            self.error('concept_does_not_exist')

    def validate_field(self):
        """
        Validate and clean the field.

        If a concept is also available, it will accessible via `self.concept`.
        """
        field = self.data.get('field')

        if not field:
            self.error('field_not_defined')

        concept = self.cleaned_data.get('concept')

        # A non-falsy concept is defined, but no cleaned concept is available.
        # The field cannot be validated
        if self.data.get('concept') and not concept:
            return

        kwargs = utils.parse_field_key(field)

        # If the concept is defined, restrict to the concept, otherwise
        # get from the entire set.
        if concept:
            queryset = concept.fields.all()
        elif 'user' in self.context:
            queryset = DataField.objects.published(user=self.context['user'])
        else:
            queryset = DataField.objects.all()

        try:
            return queryset.get(**kwargs)
        except DataField.DoesNotExist:
            if concept:
                self.error('field_does_not_exist_for_concept')
            self.error('field_does_not_exist')
        except DataField.MultipleObjectsReturned:
            if concept:
                self.error('ambiguous_field_for_concept')
            self.error('ambiguous_field')



from optparse import make_option
from django.db import transaction
from django.db.models import Q
from django.core.management.base import BaseCommand
from avocado.core import utils
from avocado.models import DataField, DataCategory, DataConcept, \
    DataConceptField
from avocado.legacy import models as legacy


class Command(BaseCommand):
    help = "Utilities for migrating Avocado 1.x metadata to the Avocado 2.x "
    "data model."

    option_list = BaseCommand.option_list + (
        make_option('-f', '--force', action='store_true', dest='force',
                    default=False, help='During a migration, force an update '
                    'to existing data.'),

        make_option('--no-input', action='store_true', dest='no_input',
                    default=False, help='Prevents user prompts and performs '
                    'the default bahavior'),

        make_option('--no-fields', action='store_false', dest='fields',
                    default=True, help='Do migrate legacy Field instances to '
                    'DataField'),

        make_option('--no-columns', action='store_false', dest='columns',
                    default=True, help='Do migrate legacy Column instances to '
                    'DataConcept'),

        make_option('--no-criteria', action='store_false', dest='criteria',
                    default=True, help='Do migrate legacy Criterion instances '
                    'to DataConcept'),

        make_option('--no-categories', action='store_false', dest='categories',
                    default=True, help='Do migrate legacy Category instances '
                    'to DataCategory'),
    )

    def _migrate_categories(self, **options):
        total_migrated = 0

        # Migrate categories without a parent first
        for lc in legacy.Category.objects.order_by('-parent').iterator():
            kwargs = {
                'name__iexact': lc.name,
            }
            # Filter by parent if one exists since categories with the same
            # name can exists as sub-categories.
            if lc.parent_id:
                kwargs['parent__name'] = lc.parent.name

            if DataCategory.objects.filter(**kwargs).exists():
                print '"{0}" already exists. Skipping...'.format(lc)
                continue

            if lc.parent_id:
                try:
                    p = DataCategory.objects.get(name=lc.parent.name)
                except DataCategory.DoesNotExist:
                    print('Parent category "{0}" does not exist for "{1}". '
                          'Skipping...'.format(lc.parent.name, lc.name))
                    continue
            else:
                p = None

            c = DataCategory(name=lc.name, parent=p, order=lc.order)
            c.save()
            print 'Migrated "{0}"'.format(c)
            total_migrated += 1

        print u'Categories migrated:\t{0}'.format(total_migrated)

    def _migrate_fields(self, **options):
        force = options.get('force')
        no_input = options.get('no_input')

        total_migrated = 0

        for lf in legacy.Field.objects.iterator():
            try:
                f = DataField.objects.get_by_natural_key(lf.app_name,
                                                         lf.model_name,
                                                         lf.field_name)
            except DataField.DoesNotExist:
                f = DataField(app_name=lf.app_name, model_name=lf.model_name,
                              field_name=lf.field_name)

            qualified_name = \
                u'({0}) {1}.{2}'.format(f.app_name, f.model_name, f.field_name)

            if f.pk and not force:
                print u'{0} already exists. Skipping...'.format(qualified_name)
                continue

            # Check if this is an orphan
            if not f.field:
                print u'{0} is orphaned. Skipping...'.format(qualified_name)
                continue

            # Map various fields
            f.name = lf.name
            f.description = lf.description
            f.keywords = lf.keywords
            f.translator = lf.translator
            f.group_id = lf.group_id

            print u'Migrated "{0}"'.format(qualified_name)

            flags = utils.get_heuristic_flags(f)
            f.__dict__.update(flags)

            # Disagreement with enumerable status
            if not no_input and f.enumerable != lf.enable_choices:
                if lf.enable_choices:
                    override = raw_input(u'"{0}" is marked as enumerable, but '
                                         'does not qualify to be enumerable. '
                                         'Override? [y/N] '
                                         .format(qualified_name))
                else:
                    override = raw_input(u'"{0}" is not marked as enumerable, '
                                         'but qualifies to be enumerable. '
                                         'Override? [y/N] '
                                         .format(qualified_name))

                if override.lower() == 'y':
                    f.enumerable = lf.enable_choices

            f.save()
            f.sites = lf.sites.all()

            total_migrated += 1

        print u'Fields migrated:\t{0}'.format(total_migrated)

    def _migrate_concept(self, model, migrate_func, **options):
        no_input = options.get('no_input')

        total_migrated = 0

        for lc in model.objects.iterator():
            field_nks = list(lc.fields.values('app_name', 'model_name',
                                              'field_name').distinct())
            field_cond = Q()

            for f in field_nks:
                field_cond = field_cond | Q(**f)

            fields = DataField.objects.filter(field_cond).distinct()

            # Mismatch of fields from new to old
            if len(fields) != len(field_nks):
                print('One or more fields mismatched for "{0}". '
                      'Skipping...'.format(lc))
                continue

            matches = DataConcept.objects.filter(name=lc.name)

            # Filter concepts by existence of fields
            for f in fields:
                matches = matches.filter(fields__app_name=f.app_name,
                                         fields__model_name=f.model_name,
                                         fields__field_name=f.field_name)

            num = len(matches)

            if num > 1:
                print('{0} have the same name and fields. '
                      'Skipping...'.format(num))
                continue

            if num == 1:
                c = matches[0]
                existing = True

                if not no_input:
                    override = True
                    while True:
                        response = raw_input(u'Match found for "{0}". '
                                             'Override? [n/Y] '.format(c))
                        if not response:
                            break
                        if response.lower() == 'n':
                            override = False
                            break
                    if not override:
                        continue
            else:
                c = DataConcept(queryable=False, viewable=False)
                existing = False

            c.name = lc.name
            c.order = lc.order
            c.published = lc.is_public

            # This looks odd, but this handles choosing the longer of the two
            # descriptions for criterion and column if both exist.
            if not c.description or lc.description and \
                    len(lc.description) > len(c.description):
                c.description = lc.description

            if lc.category:
                try:
                    kwargs = {
                        'name__iexact': lc.category.name,
                    }
                    # Filter by parent if one exists since categories with the
                    # same name can exists as sub-categories.
                    if lc.category.parent_id:
                        kwargs['parent__name'] = lc.category.parent.name
                    c.category = DataCategory.objects.get(**kwargs)
                except DataCategory.DoesNotExist:
                    pass

            # Apply migration specific function to concept from legacy concept
            migrate_func(c, lc)

            # Save for foreign key references to concept fields
            c.save()

            cfs = []

            if not existing:
                lcfs = list(lc.conceptfields.select_related('field'))

                # Dict of legacy concept fields to the new field it
                # corresponds to
                lcf_map = {}

                for lcf in lcfs:
                    for f in fields:
                        # Match and break
                        if lcf.field.natural_key() == f.natural_key():
                            lcf_map[lcf.pk] = f
                            break

                # Map fields to
                # Iterate over all legacy concept fields and create the new
                # concept fields
                for lcf in lcfs:
                    f = lcf_map[lcf.pk]
                    cfs.append(DataConceptField(concept=c, field=f,
                                                name=lcf.name,
                                                order=lcf.order))

            # Save concept fields
            for cf in cfs:
                cf.save()

            print 'Migrated "{0}"'.format(c)
            total_migrated += 1

        print '{0} migrated: {1}'.format(model.__name__, total_migrated)

    def _migrate_criteria(self, **options):
        def func(c, lc):
            c.queryable = True
        self._migrate_concept(legacy.Criterion, func, **options)

    def _migrate_columns(self, **options):
        def func(c, lc):
            c.viewable = True
            c.formatter_name = lc.html_fmtr
        self._migrate_concept(legacy.Column, func, **options)

    def handle(self, **options):
        if options.get('categories'):
            with transaction.commit_on_success():
                self._migrate_categories(**options)
        if options.get('fields'):
            with transaction.commit_on_success():
                self._migrate_fields(**options)
        if options.get('criteria'):
            with transaction.commit_on_success():
                self._migrate_criteria(**options)
        if options.get('columns'):
            with transaction.commit_on_success():
                self._migrate_columns(**options)



import warnings
from avocado.conf import dep_supported, raise_dep_error

if not dep_supported('objectset'):
    raise_dep_error('objectset')

from objectset.models import ObjectSet, SetObject  # noqa

warnings.warn('The built-in ObjectSet and SetObject classes have been '
              'removed in favor of the classes defined in the '
              'django-objectset library. Imports should be changed to '
              'import directly from objectset.models instead of from '
              'avocado.sets.models')



import inspect
from django.conf import settings
from django.utils.importlib import import_module


class AlreadyRegistered(Exception):
    pass


class NotRegistered(Exception):
    pass


class Registry(object):
    "Registry for classes or instances."
    def __init__(self, default=None, name=None, register_instance=True):
        self.register_instance = register_instance
        self._registry = {}
        self.default = None

        # Register the default class/instance
        if default:
            self.register(default, name, default=True)

    def __len__(self):
        return len(self._registry)

    def __nonzero__(self):
        return True

    def __contains__(self, key):
        return key in self._registry

    def __getitem__(self, key):
        return self._registry.get(key, self.default)

    def get(self, name):
        if name in self:
            return self[name]
        return self.default

    def register(self, obj, name=None, default=False):
        """Registers a class or instance with an optional name. The class
        name will be used if not supplied.
        """
        if default and not name:
            name = 'Default'

        if inspect.isclass(obj):
            name = name or obj.__name__
            # Create an instance if instances should be registered
            if self.register_instance:
                obj = obj()
        else:
            name = name or obj.__class__.__name__

        if default:
            self.default = self._registry[name] = obj
        elif name in self._registry:
            raise AlreadyRegistered(u'The object "{0}" is already '
                                    'registered'.format(name))
        self._registry[name] = obj

    def unregister(self, name):
        """Unregisters an object.

        **Note:** that these calls must be made in INSTALLED_APPS listed
        after the apps that already registered the class.
        """
        # Use the name of the class if passed in. Second condition checks
        # for an instance of the class.
        if inspect.isclass(name):
            name = name.__name__
        elif hasattr(name, '__class__'):
            name = name.__class__.__name__

        if name not in self._registry:
            raise NotRegistered(u'No object "{0}" registered'.format(name))
        self._registry.pop(name)

    @property
    def choices(self):
        "Returns a 2-tuple list of all registered class instance names."
        return sorted((x, x) for x in self._registry.iterkeys())


def autodiscover(module_name):
    """Simple auto-discover for looking through each INSTALLED_APPS for each
    module_name and fail silently when not found or if an error occurs.
    """
    for app in settings.INSTALLED_APPS:
        # Attempt to import the app's ``module_name``.
        try:
            import_module('{0}.{1}'.format(app, module_name))
        except:
            pass



from django.db.models import Q
from avocado.models import DataField


def get_fields_by_label(labels):
    """Constructs a DataField QuerySet given a list of labels.

    The label format can be <app>, <app.model> or <app.model.field>.
    """
    conditions = []

    for label in labels:
        labels = label.split('.')

        # Specific field
        if len(labels) == 3:
            app, model, field = labels
            conditions.append(Q(app_name=app, model_name=model,
                                field_name=field))
        # All fields for a model
        elif len(labels) == 2:
            app, model = labels
            conditions.append(Q(app_name=app, model_name=model))
        # All fields for each model in the app
        else:
            app = labels[0]
            conditions.append(Q(app_name=app))

    fields = DataField.objects.all()

    if conditions:
        q = Q()
        for x in conditions:
            q = q | x
        fields = fields.filter(q).distinct()

    return fields



from . import datacontext as datacontext_parser
from . import dataview as dataview_parser


class Node(object):
    datacontext_node = None
    dataview_node = None

    def __init__(self, datacontext_node=None, dataview_node=None, **context):
        self.datacontext_node = datacontext_node
        self.dataview_node = dataview_node

    def apply(self, queryset=None, distinct=True, include_pk=True):
        queryset = \
            self.datacontext_node.apply(queryset=queryset, distinct=distinct)
        return \
            self.dataview_node.apply(queryset=queryset, include_pk=include_pk)


def validate(attrs, **context):
    if not attrs:
        return

    datacontext_attrs = attrs.get('context', {})
    dataview_attrs = attrs.get('view', {})

    ret_attrs = {}

    ret_attrs['context'] = \
        datacontext_parser.validate(datacontext_attrs, **context)

    ret_attrs['view'] = dataview_parser.validate(dataview_attrs, **context)

    return ret_attrs


def parse(attrs, tree=None, **context):
    if not attrs:
        return Node(**context)

    datacontext_attrs = attrs.get('context', {})
    datacontext_node = \
        datacontext_parser.parse(datacontext_attrs, tree=tree, **context)

    dataview_attrs = attrs.get('view', {})
    dataview_node = \
        dataview_parser.parse(dataview_attrs, tree=tree, **context)

    return Node(datacontext_node, dataview_node, **context)



import logging
from django.db import models
from django.db.models import Q
from django.db import transaction
from django.conf import settings
from django.db.models.manager import ManagerDescriptor
from django.core.exceptions import ImproperlyConfigured
from avocado.conf import OPTIONAL_DEPS, requires_dep
from avocado.core.managers import PublishedManager, PublishedQuerySet


logger = logging.getLogger(__name__)


class DataSearchMixin(models.Manager):
    @requires_dep('haystack')
    def search(self, content, queryset=None, max_results=None, partial=False,
               using=None):
        from haystack.query import RelatedSearchQuerySet
        from haystack.inputs import AutoQuery
        # Limit to the model bound to this manager, e.g. DataConcept.
        # `load_all` ensures a single database hit when loading the objects
        # that match
        sqs = RelatedSearchQuerySet().models(self.model).load_all()

        # If a non-default backend is being used, set which backend is to
        # be used.
        if using is not None:
            sqs = sqs.using(using)

        if partial:
            # Autocomplete only works with N-gram fields
            sqs = sqs.autocomplete(text_auto=content)
        else:
            # Automatically handles advanced search syntax for negations and
            # quoted strings
            sqs = sqs.filter(text=AutoQuery(content))

        if queryset is not None:
            sqs = sqs.load_all_queryset(self.model, queryset)

        if max_results:
            return sqs[:max_results]

        return sqs


class DataFieldQuerySet(PublishedQuerySet):
    def published(self, user=None, perm='avocado.view_datafield'):
        """Fields can be restricted to one or more sites, so the published
        method is extended to support filtering by site.
        """
        published = super(DataFieldQuerySet, self).published()

        # Remove internal
        published = published.exclude(internal=True)

        # All published concepts associated with the current site
        # (or no site)
        sites = Q(sites=None) | Q(sites__id=settings.SITE_ID)
        published = published.filter(sites)

        if user:
            if not OPTIONAL_DEPS['guardian']:
                raise ImproperlyConfigured('django-guardian must installed '
                                           'for object-level permissions.')
            from guardian.shortcuts import get_objects_for_user
            published = get_objects_for_user(user, perm, published)
        return published.distinct()


class DataConceptQuerySet(PublishedQuerySet):
    def published(self, user=None, perm='avocado.view_datafield'):
        """Concepts can be restricted to one or more sites, so the published
        method is extended to support filtering by site. In addition, concepts
        should not be visible if their associated fields are not all available.
        Also, concepts with an unpublished category are not visible. Finally,
        concepts with no fields are not considered visible.
        """
        published = super(DataConceptQuerySet, self).published()

        # Remove internal
        published = published.exclude(internal=True)

        # All published concepts associated with the current site
        # (or no site)
        sites = Q(sites=None) | Q(sites__id=settings.SITE_ID)
        published = published.filter(sites)

        # All published concepts with a non-empty category must have the
        # category published as well in order for it to be visible.
        category_published = Q(category=None) | Q(category__published=True)
        published = published.filter(category_published)

        # Concepts that contain at least one unpublished or archived datafield
        # are removed from the set to prevent exposing unprepared data
        from avocado.models import DataField
        fields_q = Q(archived=True) | Q(published=False)
        if user:
            if not OPTIONAL_DEPS['guardian']:
                raise ImproperlyConfigured('django-guardian must installed '
                                           'for object-level permissions.')
            from guardian.shortcuts import get_objects_for_user
            # If a user is specified, they must also have a permission for
            # accessing the data fields. All data fields that the user does
            # NOT have access to must also be removed from the set.
            restricted_fields = DataField.objects\
                .exclude(pk__in=get_objects_for_user(user, perm))
            fields_q = fields_q | Q(pk__in=restricted_fields)

        shadowed = DataField.objects.filter(fields_q)
        concepts = published.exclude(fields__in=shadowed) \
            .exclude(fields=None).distinct()
        return concepts


class DataFieldManagerDescriptor(ManagerDescriptor):
    """Manager descriptor customized to allow for `f.objects` which returns
    a QuerySet of the underlying model objects which DataField represents.
    """
    def __init__(self, manager):
        self.manager = manager

    def __get__(self, instance, type=None):
        if instance and isinstance(instance.field, models.AutoField):
            return instance.model.objects.all()
        return super(DataFieldManagerDescriptor, self).__get__(instance, type)


class DataFieldManager(PublishedManager, DataSearchMixin):
    "Manager for the `DataField` model."

    def contribute_to_class(self, model, name):
        "Override to use custom manager descriptor."
        super(DataFieldManager, self).contribute_to_class(model, name)
        setattr(model, name, DataFieldManagerDescriptor(self))

    def get_query_set(self):
        return DataFieldQuerySet(self.model, using=self._db)

    def get_by_natural_key(self, app_name, model_name=None, field_name=None):
        queryset = self.get_query_set()
        if type(app_name) is int:
            return queryset.get(id=app_name)

        # Extract attributes from model field instance
        if isinstance(app_name, models.Field):
            field = app_name
            opts = field.model._meta
            app_name = opts.app_label
            model_name = opts.module_name
            field_name = field.name

        keys = ['app_name', 'model_name', 'field_name']

        if type(app_name) is list:
            values = app_name
        elif type(app_name) and '.' in app_name:
            values = app_name.split('.')
        else:
            values = [app_name, model_name, field_name]
        return queryset.get(**dict(zip(keys, values)))


class DataConceptManager(PublishedManager, DataSearchMixin):
    "Manager for the `DataConcept` model."
    def get_query_set(self):
        return DataConceptQuerySet(self.model, using=self._db)

    @transaction.commit_on_success
    def create_from_field(self, field, save=True, **kwargs):
        """Derives a DataConcept from this DataField's descriptors. Additional
        keyword arguments can be passed in to customize the new DataConcept
        object. The DataConcept can also be optionally saved by setting the
        `save` flag.
        """
        for key, value, in field.descriptors.iteritems():
            kwargs.setdefault(key, value)

        concept = self.model(**kwargs)

        if save:
            from avocado.models import DataConceptField
            concept.save()
            cfield = DataConceptField(field=field, concept=concept)
            concept.concept_fields.add(cfield)
        return concept


class DataCategoryManager(PublishedManager):
    "Manager for the `DataCategory` model."


class DataClassBaseManager(models.Manager):
    def get_default_template(self):
        try:
            return self.get_query_set().get(default=True, template=True)
        except self.model.DoesNotExist:
            pass


class DataViewManager(DataClassBaseManager):
    "Manager for the `DataView` model."


class DataContextManager(DataClassBaseManager):
    "Manager for the `DataContext` model."


class DataQueryManager(DataClassBaseManager):
    "Manager for the `DataQuery` model."



import re
import os
import sys
import tempfile
import logging
from django.core import serializers, management
from django.core.exceptions import ImproperlyConfigured
from django.db import models, transaction, router, DEFAULT_DB_ALIAS, \
    DatabaseError, IntegrityError
from avocado.models import DataField, DataConcept, DataConceptField, \
    DataCategory

FIXTURE_FORMAT = 'json'
FIXTURE_FILENAME_RE = re.compile(r'^[0-9]{4}_[0-9a-zA-Z_]+(\.json)$')

MIGRATION_MODEL_LABELS = ('avocado.datafield', 'avocado.dataconcept',
                          'avocado.dataconceptfield', 'avocado.datacategory')

log = logging.getLogger(__name__)


def _check_app():
    from avocado.conf import settings
    app_name = settings.METADATA_MIGRATION_APP
    if not app_name or not models.get_app(app_name.split('.')[-1]):
        raise ImproperlyConfigured('{0} is not a valid app name. The '
                                   'backup/migration API requires '
                                   'METADATA_MIGRATION_APP to be a valid '
                                   'app name to be used.'
                                   .format(settings.METADATA_MIGRATION_APP))


def _fixture_filenames(dirname):
    filenames = []
    for f in os.listdir(dirname):
        if FIXTURE_FILENAME_RE.match(os.path.basename(f)):
            filenames.append(f)
    filenames.sort()
    return filenames


def next_fixture_name(name, dirname):
    filenames = _fixture_filenames(dirname)
    if not filenames:
        version = '0001'
    else:
        version = unicode(int(filenames[-1][:4]) + 1).zfill(4)
    return u'{0}_{1}'.format(version, name)


def full_fixture_path(name):
    return u'{0}.{1}'.format(os.path.join(get_fixture_dir(), name),
                             FIXTURE_FORMAT)


def get_fixture_dir():
    from avocado.conf import settings
    _check_app()

    fixture_dir = settings.METADATA_FIXTURE_DIR
    if fixture_dir:
        return fixture_dir
    app = models.get_app(settings.METADATA_MIGRATION_APP.split('.')[-1])
    if hasattr(app, '__path__'):
        # It's a 'models/' subpackage
        app_dir = os.path.dirname(os.path.dirname(app.__file__))
    else:
        # It's a models.py module
        app_dir = os.path.dirname(app.__file__)
    return os.path.join(app_dir, 'fixtures')


def create_fixture(name, using=DEFAULT_DB_ALIAS, silent=False):
    "Dumps data in the fixture format optionally to a particular path."
    if os.path.isabs(name):
        fixture_path = name
    else:
        fixture_path = full_fixture_path(name)
    with open(fixture_path, 'w') as fout:
        management.call_command('dumpdata', *MIGRATION_MODEL_LABELS,
                                database=using, stdout=fout)
    if not silent:
        log.info(u'Created fixture {0}'.format(name))


def create_temp_fixture(*args, **kwargs):
    "Creates a fixture to a temporary file."
    fh, path = tempfile.mkstemp()
    create_fixture(path, *args, **kwargs)
    return path


def load_fixture(name, using=DEFAULT_DB_ALIAS):
    """Progammatic way to load a fixture given some path. This does not
    assume the path is a fixture within some app and assumes a full path.
    """
    if os.path.isabs(name):
        fixture_path = name
    else:
        fixture_path = full_fixture_path(name)

    with open(fixture_path) as fixture:
        objects = serializers.deserialize(FIXTURE_FORMAT, fixture, using=using)

        with transaction.commit_manually(using):
            for obj in objects:
                if router.allow_syncdb(using, obj.object.__class__):
                    try:
                        obj.save(using=using)
                    except (DatabaseError, IntegrityError), e:
                        transaction.rollback(using)
                        msg = u'Could not load {0}.{1}(pk={2}): {3}'.format(
                            obj.object._meta.app_label,
                            obj.object._meta.object_name, obj.object.pk, e)
                        raise e.__class__, e.__class__(msg), sys.exc_info()[2]
            transaction.commit(using)
    log.info(u'Loaded data from fixture {0}'.format(name))


def delete_metadata(using=DEFAULT_DB_ALIAS):
    "Deletes all metadata in the target database."
    DataConceptField.objects.using(using).delete()
    DataConcept.objects.using(using).delete()
    DataField.objects.using(using).delete()
    DataCategory.objects.using(using).delete()
    log.debug(u'Metadata deleted from database "{0}"'.format(using))


def safe_load(name, backup_path=None, using=DEFAULT_DB_ALIAS):
    """Creates a backup of the current state of the metadata, attempts to load
    the new fixture and falls back to the backup fixture if the load fails for
    any reason.
    """
    _check_app()

    with transaction.commit_manually(using):
        # Create the backup fixture
        if backup_path:
            create_fixture(os.path.abspath(backup_path), using=using,
                           silent=True)
        else:
            backup_path = create_temp_fixture(using=using, silent=True)
        log.info(u'Backup fixture written to {0}'.format(os.path.abspath(
            backup_path)))
        delete_metadata(using=using)
        try:
            load_fixture(name, using=using)
        except (DatabaseError, IntegrityError):
            transaction.rollback(using)
            log.error(u'Fixture load failed, reverting from backup: {0}'
                      .format(backup_path))
            load_fixture(backup_path, using=using)
            raise
        transaction.commit(using)
    return backup_path



import warnings
from haystack import indexes
from avocado.conf import settings
from avocado.models import DataConcept, DataField

# Warn if either of the settings are set to false
if not getattr(settings, 'CONCEPT_SEARCH_ENABLED', True) or \
        not getattr(settings, 'FIELD_SEARCH_ENABLED', True):
    warnings.warn('CONCEPT_SEARCH_ENABLED and FIELD_SEARCH_ENABLED have been '
                  'deprecated due to changes in Haystack 2.x API. To exclude '
                  'an index from being discovered, add the path to the class '
                  'to EXCLUDED_INDEXES in the appropriate '
                  'HAYSTACK_CONNECTIONS entry in settings.')


class DataIndex(indexes.SearchIndex):
    text = indexes.CharField(document=True, use_template=True)
    text_auto = indexes.EdgeNgramField(use_template=True)

    def index_queryset(self, using=None):
        return self.get_model().objects.published()

    def load_all_queryset(self):
        return self.index_queryset()


class DataConceptIndex(DataIndex, indexes.Indexable):
    def get_model(self):
        return DataConcept


class DataFieldIndex(DataIndex, indexes.Indexable):
    def get_model(self):
        return DataField



import jsonfield
from django.db import models
from modeltree.tree import trees
from . import oldparsers as parsers


def _sql_string(queryset):
    sql, params = queryset.query.sql_with_params()
    return sql % tuple([repr(str(x)) for x in params])


class AbstractDataContext(models.Model):
    """JSON object representing one or more data field conditions. The data may
    be a single condition, an array of conditions or a tree stucture.

    This corresponds to the `WHERE` statements in a SQL query.
    """
    json = jsonfield.JSONField(null=True, blank=True, default=dict,
                               validators=[parsers.datacontext.validate])
    count = models.IntegerField(null=True, db_column='_count')
    tree = models.CharField(max_length=100, null=True, blank=True)

    def __init__(self, *args, **kwargs):
        if args and isinstance(args[0], dict):
            if 'json' in kwargs:
                raise TypeError("{0}.__init__() got multiple values for "
                                "keyword argument 'json'"
                                .format(self.__class__.__name__))
            args = list(args)
            kwargs['json'] = args.pop(0)
        super(AbstractDataContext, self).__init__(*args, **kwargs)

    class Meta(object):
        abstract = True

    def _combine(self, other, operator):
        if not isinstance(other, self.__class__):
            raise TypeError('Other object must be a DataContext instance')
        cxt = self.__class__()
        cxt.user_id = self.user_id or other.user_id
        if self.json and other.json:
            cxt.json = {
                'type': operator,
                'children': [
                    {'composite': self.pk},
                    {'composite': other.pk}
                ]
            }
        elif self.json:
            cxt.json = {'composite': self.pk}
        elif other.json:
            cxt.json = {'composite': other.pk}
        return cxt

    def __and__(self, other):
        return self._combine(other, 'and')

    def __or__(self, other):
        return self._combine(other, 'or')

    @property
    def model(self):
        "The model this context represents with respect to the count."
        if self.tree in trees:
            return trees[self.tree].root_model
        return trees.default.root_model

    @classmethod
    def validate(cls, attrs, **context):
        "Validate `attrs` as a context."
        return parsers.datacontext.validate(attrs, **context)

    def parse(self, tree=None, **context):
        "Returns a parsed node for this context."
        return parsers.datacontext.parse(self.json, tree=tree, **context)

    def apply(self, queryset=None, tree=None, **context):
        "Applies this context to a QuerySet."
        if tree is None and queryset is not None:
            tree = queryset.model
        return self.parse(tree=tree, **context).apply(queryset=queryset)

    def language(self, tree=None, **context):
        return self.parse(tree=tree, **context).language

    def sql(self, *args, **kwargs):
        """Returns the SQL query string representative of this context.

        This takes the same arguments as `apply()`.
        """
        return _sql_string(self.apply(*args, **kwargs))


class AbstractDataView(models.Model):
    """JSON object representing one or more data field conditions. The data may
    be a single condition, an array of conditions or a tree stucture.

    This corresponds to the `SELECT` and `ORDER BY` statements in a SQL query.
    """
    json = jsonfield.JSONField(null=True, blank=True, default=dict,
                               validators=[parsers.dataview.validate])

    def __init__(self, *args, **kwargs):
        if args and isinstance(args[0], dict):
            if 'json' in kwargs:
                raise TypeError("{0}.__init__() got multiple values for "
                                "keyword argument 'json'"
                                .format(self.__class__.__name__))
            args = list(args)
            kwargs['json'] = args.pop(0)
        super(AbstractDataView, self).__init__(*args, **kwargs)

    class Meta(object):
        abstract = True

    @classmethod
    def validate(cls, attrs, **context):
        "Validates `attrs` as a view."
        return parsers.dataview.validate(attrs, **context)

    def parse(self, tree=None, **context):
        "Returns a parsed node for this view."
        return parsers.dataview.parse(self.json, tree=tree, **context)

    def apply(self, queryset=None, tree=None, include_pk=True, **context):
        "Applies this context to a QuerySet."
        if tree is None and queryset is not None:
            tree = queryset.model
        return self.parse(tree=tree, **context) \
            .apply(queryset=queryset, include_pk=include_pk)

    def sql(self, *args, **kwargs):
        """Returns the SQL query string representative of this view.

        This takes the same arguments as `apply()`.
        """
        return _sql_string(self.apply(*args, **kwargs))


class AbstractDataQuery(models.Model):
    """
    JSON object representing a complete query.

    The query is constructed from a context(providing the 'WHERE' statements)
    and a view(providing the 'SELECT' and 'ORDER BY" statements). This
    corresponds to all the statements of the SQL query to dictate what info
    to retrieve, how to filter it, and the order to display it in.
    """
    context_json = jsonfield.JSONField(
        null=True, blank=True, default=dict,
        validators=[parsers.datacontext.validate])
    view_json = jsonfield.JSONField(
        null=True, blank=True, default=dict,
        validators=[parsers.dataview.validate])

    # The count when just the context is applied
    distinct_count = models.IntegerField(null=True)
    # The count when the context and the view is applied
    record_count = models.IntegerField(null=True)
    tree = models.CharField(max_length=100, null=True, blank=True)

    def __init__(self, *args, **kwargs):
        if args and isinstance(args[0], dict):
            if 'context_json' in kwargs:
                raise TypeError("{0}.__init__() got multiple values for "
                                "keyword argument 'context_json'"
                                .format(self.__class__.__name__))

            if 'view_json' in kwargs:
                raise TypeError("{0}.__init__() got multiple values for "
                                "keyword argument 'view_json'"
                                .format(self.__class__.__name__))

            kwargs['context_json'] = args[0].get('context', None)
            kwargs['view_json'] = args[0].get('view', None)

        super(AbstractDataQuery, self).__init__(*args, **kwargs)

    class Meta(object):
        abstract = True
        verbose_name_plural = 'data queries'

    @property
    def context(self):
        return AbstractDataContext(json=self.context_json)

    @property
    def view(self):
        return AbstractDataView(json=self.view_json)

    @property
    def json(self):
        return {
            'context': self.context_json,
            'view': self.view_json
        }

    @property
    def model(self):
        "The model this query represents with respect to the counts."
        if self.tree in trees:
            return trees[self.tree].root_model
        return trees.default.root_model

    @classmethod
    def validate(cls, attrs, **context):
        "Validates `attrs` as a query."
        return parsers.dataquery.validate(attrs, **context)

    def parse(self, tree=None, **context):
        "Returns a parsed node for this query."
        json = {
            'context': self.context_json,
            'view': self.view_json,
        }
        return parsers.dataquery.parse(json, tree=tree, **context)

    def apply(self, queryset=None, tree=None, distinct=True, include_pk=True,
              **context):
        "Applies this context to a QuerySet."
        if tree is None and queryset is not None:
            tree = queryset.model
        return self.parse(tree=tree, **context) \
            .apply(queryset=queryset, distinct=distinct, include_pk=include_pk)

    def sql(self, *args, **kwargs):
        """Returns the SQL query string representative of this query.

        This takes the same arguments as `apply()`.
        """
        return _sql_string(self.apply(*args, **kwargs))



import math
import random
from collections import defaultdict


def std_dev(values):
    """
    Computes the standard deviation of the 'values' list.

    TODO: Consider adding support for axes as in numpy and letting this method
    accept nested lists:
        http://docs.scipy.org/doc/numpy/reference/generated/numpy.std.html

    Arguments:
        values: list
            A list of numbers to compute the standard deviation of.

    Returns:
        The standard deviation of the elements in the 'values' list.
    """
    # Compute the mean
    mean = sum(values) / float(len(values))

    # Compute the square difference of all the values
    square_differences = [(v - mean)**2 for v in values]

    return math.sqrt(sum(square_differences) / float(len(square_differences)))


def is_iterable(obj):
    """
    Returns True if 'obj' is iterable and False otherwise.

    We check for __iter__ and for __getitem__ here because an iterable type
    must, by definition, define one of these attributes but is not required or
    guarenteed to define both of them. For more information on iterable types
    see:

        http://docs.python.org/2/glossary.html#term-iterable
        http://docs.python.org/2/library/functions.html#iter
    """
    return hasattr(obj, '__iter__') or hasattr(obj, '__getitem__')


def divide_by_scalar(lst, s):
    """
    Divides each element in 'lst' by the scalar 's'.

    Returns a new list with each element of the new list equal to the element
    at the same position in 'lst' divided by 's'.
    """
    return [l / float(s) for l in lst]


def divide_lists(lst_numer, lst_denom):
    """
    Divides each element of the nested list 'lst_numer' by 'lst_denom'.

    The division is done by taking each element of 'lst_numer' and for each
    index of that element, dividing the item at that index by the item at the
    same index of 'lst_denom'. See example below:

        >>> numer = [[1., 2.],
        ...          [3., 4.]]
        >>> denom = [0.5, 0.5]
        >>> divide_lists(numer, denom)
        [[2., 4.], [6., 8.]]

    NOTE: It is assumed that each element of 'lst_numer' has the same length
    as 'lst_denom' as shown in the dimensions of the arguments below.

    Arguments:
        lst_numer: nested list of list(dimensions N x M)
        lst_denom: list of denominators(dimensions 1 x M)

    Returns:
        A new list formed by dividing each element of 'lst_numer' by
        'lst_denom' according to the division process described above.
    """
    indexes = range(len(lst_denom))
    return [[n[i] / float(lst_denom[i]) for i in indexes] for n in lst_numer]


def normalize(points):
    """
    Normalizes a set of points on a per dimension basis.

    Each dimension is divided by its standard deviation accross all points.

    NOTE: If all the points are the same along any dimension, the resulting
    normalized value of all points along that dimension will be 0 because the
    standard deviation of a collection of equal values is 0 and we can't
    realistically divide by 0 so we instead set the normalized value to 0.

    Arguments:
        points: list of points
            Each row of the supplied list is a point and each column of those
            rows is a dimension of that point as shown below.

            #               d0  d1  d2
            >>> points = [[ 1., 2., 3.],  #  p0
            ...           [ 4., 5., 6.],  #  p1
            ...           [ 7., 8., 9.]]  #  p2

            Single dimension point list:
            #             d0
            >>> points = [1.,   #p0
            ...           2.,   #p1
            ...           3.,   #p2
            ...           4.]   #p3

            Equal points list:
            >>> points = [1., 1., 1.]
            [0., 0., 0.]

    Returns:
        The values in 'points' scaled by the standard deviation along each
        dimension.
    """
    # Check for a single dimension list. This check assumes that if the first
    # element is not a list then all elements are non-list and the list is
    # single dimension. If the list is single dimension just divide all the
    # points by the standard deviation and return that as the normalized list.
    if len(points) > 0 and not is_iterable(points[0]):
        std = std_dev(points)
        if std == 0:
            return [0] * len(points)
        else:
            return divide_by_scalar(points, std_dev(points))

    # Organize the points list as a list where each row is a list of values of
    # the same dimension.
    dimensions = zip(*points)

    # Compute the standard deviation of each dimension.
    std = [std_dev(d) for d in dimensions]

    # Divide all the point dimensions by the standard deviation of that
    # dimension over all points.
    if 0 in std:
        norm_points = []
        for point in points:
            norm_point = []
            for p, s in zip(point, std):
                if s == 0:
                    norm_point.append(0.0)
                else:
                    norm_point.append(p / s)

            norm_points.append(norm_point)

        return norm_points
    else:
        return divide_lists(points, std)


def is_nested(points):
    """
    Returns True if points is a nested iterable and False if not.

    A nested iterable is an iterable that is made of iterable elements. See
    the example below:

        >>> p = [[1, 2, 3],
        ...      [4, 5, 6]]
        >>> is_nested(p)
        True
        >>> p = [10, 20, 30]
        >>> is_nested(p)
        False
        >>> p = [[10], [20], [30]]
        >>> is_nested(p)
        True

    NOTE: This method only checks the first element of the iterable 'points'
    to make the determination of "nestedness". It is assumed that all elements
    of 'points' are either iterable or not iterable, that is, they match the
    property of the first element.

    """
    if is_iterable(points) and len(points) > 0:
        return is_iterable(points[0])

    return False


def get_dimension(points):
    """
    Returns the dimension of the 'points' list.

    The dimension is considered the number of dimensions(or columns) that make
    up the points in the list. For more details, see the examples below:

        >>> p = [[1, 2, 3],
        ...      [4, 5, 6]]
        >>> get_dimension(p)
        3
        >>> p = [10, 20, 30]
        >>> get_dimension(p)
        1

    This method checks that all points have the same dimension and will raise
    a ValueError if all the points do not share the same dimension.

    Arguments:
        points: list of points of any dimension

    Returns:
        The dimension of the point list(aka the number of columns) or raises a
        ValueError if the number of dimensions is not consistent over all
        points in 'points'.
    """

    if is_iterable(points[0]):
        dimension = len(points[0])
    else:
        dimension = 1

    i_dimension = -1
    for i in range(len(points)):
        if is_iterable(points[i]):
            i_dimension = len(points[i])
        else:
            i_dimension = 1

        if i_dimension != dimension:
            raise ValueError("Points must have the same number of dimensions.")

    return dimension


def sqr_euclidean_dist(point1, point2):
    """
    Calculates the square Euclidean distance between all dimension of 2 points.

    The returned value is a list of distances representing the squared
    Euclidean distance along each dimension of the two points. This is shown
    in the example below:

        >>> p1 = [1, 2, 3]
        >>> p2 = [3, 6, 9]
        >>> sqr_euclidean_dist(p1, p2)
        [4, 14, 36]

        >>> o1 = 5.4
        >>> o2 = 2.4
        >>> sqr_euclidean_dist(p1, p2)
        9.000000000000004

    NOTE: This method assumes 'point1' and 'point2' are both numbers or lists
    of equal length. No verification is done to make sure that both have the
    same type or that the list lengths are equal.

    Arguments:
        point1, point2: number or list of numbers

    Returns:
        The square Euclidean distance along each dimensions of the points.
    """
    if is_iterable(point1):
        return [(p1 - p2) ** 2 for p1, p2 in zip(point1, point2)]

    return (point1 - point2) ** 2


def index_of_min(values):
    """
    Finds and returns the index of the smallest item in the 'values' list.

    If multiple values are all equal to the minimum value then the one with
    the smallest index is returned.
    """
    min_value = float('Inf')
    min_index = 0

    for i in range(len(values)):
        if values[i] < min_value:
            min_value = values[i]
            min_index = i

    return min_index


def compute_clusters(points, centroids):
    """
    Computes cluster assignment and distance to cluster centroid for 'points'.

    Runs a vector quantization algorithm on the points using 'centroids' as
    the training set. This method can be considered the encoder part of the
    vector quantization algorithm where the codebook that is generated contains
    the cluster(centroid index) that each point belongs to and the distance
    between each point and its cluster centroid. The distance referred to
    above is the Euclidean distance.

    For some general info on vector quantization theory see:
        http://www.oocities.org/stefangachter/VectorQuantization/chapter_1.htm
        http://www.data-compression.com/vq.html

    Returns:
        clusters: A list such that clusters[i] is the index of the centroid
                  of the cluster that the ith point is a member of. This list
                  will have the same length as 'points'.
        distances: A list such that distances[i] represents the Euclidean
                   distance between the ith point and the centroid of the
                   cluster it is a member of. This list will have the same
                   length as 'points'.
    """
    if len(points) < 1:
        raise ValueError("cluster requires at least one point.")
    if len(centroids) < 1:
        raise ValueError("cluster requires at least one centroid.")

    d = get_dimension(points)
    n = len(points)
    cluster_indexes = range(len(centroids))

    if d != get_dimension(centroids):
        raise ValueError('Points and centroids must have the same '
                         'dimension(found {0} and {1} respectively)'.format(
                             d, get_dimension(centroids)))

    clusters = [0] * n
    min_distances = [0] * n
    for i in range(n):
        # Compute the squared Euclidean distance between this point and each
        # of the centroids.
        distances = [sqr_euclidean_dist(points[i], centroids[j])
                     for j in cluster_indexes]

        # Sum across all dimensions of each distance measure. If we only have
        # one dimension then this is simply the distance list itself.
        if d == 1 and not is_nested(centroids):
            distance_totals = distances
        else:
            distance_totals = \
                [sum(distances[j]) for j in range(len(distances))]

        # Find the index containing the minimum distance.
        clusters[i] = index_of_min(distance_totals)

        min_distances[i] = distance_totals[clusters[i]]

    # Do the square root now to get the Euclidean distance. We don't do this
    # in the loop so that we can save time by not taking the square root of
    # non-minimal distances.
    return clusters, [math.sqrt(dist) for dist in min_distances]


def dimension_mean(points):
    """
    Calculates the mean of the points along each dimension.

    If the points are of a single dimension, 'points' will just be a list of
    numbers and the mean of the list itself will be returned. For more
    details, see the example below:

        >>> p = [[1.0, 0.0],
        ...      [0.5, 2.0]]
        >>> dimension_mean(p)
        [0.75, 1.0]

        >>> p = [1., 3., 5., 10.]
        >>> dimension_mean(p)
        4.75
    """
    if get_dimension(points) == 1 and not is_nested(points):
        return sum(points) / float(len(points))

    # Organize the points list as a list where each row is a list of values
    # of the same dimension.
    dimensions = zip(*points)

    return [sum(d) / float(len(d)) for d in dimensions]


def kmeans(points, k_or_centroids, threshold=1e-5):
    """
    Runs the k-means algorithm on the points for k clusters.

    This method runs the k-means clustering algorithm assigning the 'points'
    to k clusters. k is determined in one of two ways depending on what is
    provided in 'k_or_centroids'. If 'k_or_centroids' is a list, then the
    contents are used as the centroids and k is the length of that list. If
    'k_or_centroids' is not a list of centroids it is assumed to be an integer
    to be used as k and the initial centroids are then k random points chosen
    from 'points'.

    See:
        computer_clusters()

    Returns:
        centroids: list of centroids of the clusters found during execution
        distance: the average mean distance between all points and the
                  centroid of the cluster they are a member of
    """
    # TODO: Add support for iterations if k is supplied and ignore iterations
    # if initial centroids are provided.
    if len(points) < 1:
        raise ValueError("points must contain at least 1 point.")

    if is_iterable(k_or_centroids):
        k = len(k_or_centroids)
        initial_centroids = k_or_centroids
    else:
        k = k_or_centroids
        initial_centroids = [p for p in random.sample(points, k)]

    if k < 1:
        raise ValueError("k must be >= 1.")

    centroids = list(initial_centroids)
    mean_difference = float('Inf')
    previous_mean_distance = None
    while mean_difference > threshold:
        num_centroids = len(centroids)

        # Determine cluster membership and distances for all points given the
        # current centroids.
        centroid_indexes, distances = compute_clusters(points, centroids)

        # Compute the mean distance of all points to their corresponding
        # cluster centroid.
        mean_distance = sum(distances) / float(len(distances))

        # Compute the difference in mean distance between this clustering step
        # and the last one.
        if previous_mean_distance is not None:
            mean_difference = previous_mean_distance - mean_distance

        # The following is the update step of the k-means algorithm where the
        # centroid position changes based on the points in each cluster. We
        # can safely ignore this step if the have reached the threshold for
        # minimum difference in mean distance.
        if mean_difference > threshold:
            for i in range(num_centroids):
                # Get all the points that are currently members of this
                # centroid's cluster.
                members = \
                    [p for c, p in zip(centroid_indexes, points) if c == i]

                # If the cluster has any points in it then update the centroid
                # of that cluster to be a point represented by the mean of all
                # points in that cluster along each dimension.
                if len(members) > 0:
                    centroids[i] = dimension_mean(members)
                else:
                    centroids[i] = None

            # Remove centroids of empty clusters.
            centroids = \
                [centroids[i] for i in range(len(centroids)) if centroids[i]]

        # Store this mean distance so we can access it in the next loop and
        # diff against this iterations mean distance.
        previous_mean_distance = mean_distance

    return centroids, previous_mean_distance


def find_outliers(points, outlier_threshold=3, normalized=True):
    """
    Finds and returns outliers in the 'points'.

    Outliers are those items in the 'points' that have a normalized distance
    that is at least 'outlier_threshold' away from the the calculated center
    of the population defined in 'points'. The normalized distance of a
    point is the distance from the point to its cluster's centroid divided by
    the mean distance of all point-to-centroid distances.

    Arguments:
        points: list of n-dimensional points
            An NxM list of lists defining the population to check for outliers
            where M is the number of dimensions per point and N is the number
            of points.
        outlier_threshold: float
            Used to define outliers. Outliers are points with normalized
            distances greater than this threshold.
        normalized: bool
            Value indicating whether 'points' have already been normalized or
            not. Passing False for this argument will result in the original
            'points' being normalized before the outlier search starts.

    Returns:
        The indexes of the outliers in 'points'.
    """
    if not normalized:
        points = normalize(points)

    d = get_dimension(points)

    # Organize the points list as a list where each row is a list of values of
    # the same dimension.
    if is_nested(points):
        dimensions = zip(*points)

    # Compute the mid-point index and construct the centroid by taking the
    # midpoint of the sorted list in each dimension.
    midpoint_index = (len(points) - 1) / 2

    if d == 1 and not is_nested(points):
        centroids = [sorted(points)[midpoint_index]]
    else:
        centroids = [[sorted(dim)[midpoint_index] for dim in dimensions]]

    # Calculate the distance from every point to the centroid.
    _, distances = compute_clusters(points, kmeans(points, centroids)[0])

    mean_distance = sum(distances) / float(len(distances))

    return [i for i, distance in enumerate(distances)
            if (mean_distance > 0 and
                (distance / mean_distance) >= outlier_threshold)]


def kmeans_optm(points, k=None, outlier_threshold=3):
    """
    Execute k-means clustering(for finding centroid) on, compute the clusters
    for, and run outlier detection on the 'points' population.

    Returns a dictionary containing the following data:
        'centroids': The k centers of the k clusters found by running the
                     k-means algorithm.
        'indexes':   An ordered list mapping each point in 'points' to the
                     centroid of the cluster it falls in.
        'distances': A list of distances between each point and the centroid
                     of its cluster.
        'outliers':  A list of outlier indexes in the 'points' list. This is
                     only returned if an 'outlier_threshold' is specified. If
                     outliers aren't required, set outlier_threshold=None.

    If 'k' is not supplied it will be calculated given the following formula:

        k = sqrt(n / 2)

    where n is the length of the 'points' list. The initialization method is
    optimized to find the mid-points of each cluster from a sorted list of
    points. This ultimately reduces the number of iterations required during
    clustering due to increased likelihood of convergence.

    Referenes:
        Number of clusters - http://en.wikipedia.org/wiki/Determining_the_number_of_clusters_in_a_data_set#Rule_of_thumb # noqa
        Improved k-means initial centroids - http://www.ijcsit.com/docs/vol1issue2/ijcsit2010010214.pdf # noqa

    See:
        find_outliers()

    Arguments:
        points: list of n-dimensional points
        k: int or None
            The number of clusters to calculate. If 'k' is not supplied, it
            will be calculated according to the formula descibed above.
        outlier_threshold: float
            Used to define outliers. Outliers are points with normalized
            distances greater than this threshold.
    Returns:
        A dictionary containing the centroids, indexes of each point's
        centroid, distances from each point to its centroid and optionally a
        list of indexes of the outliers in 'points'. More detailed info on
        the return values are included above.
    """
    # If an outlier threshold is defined, remove outliers relative to the
    # population before running k-means clustering.
    if outlier_threshold:
        outliers = find_outliers(points, outlier_threshold, False)
        points = [p for i, p in enumerate(points) if i not in outliers]
    else:
        outliers = []

    d = get_dimension(points)
    n = len(points)

    # Compute the number of clusters if it wasn't specified.
    k = k or int(math.sqrt(n / 2))

    # Manually calculate the standard deviation for each dimension of the
    # point list in order to de-normalize the centroids later. Otherwise, the
    # centroid would be relative to the normalized dimensions rather than the
    # original.
    if d == 1 and not is_nested(points):
        std = std_dev(points)
    else:
        # Organize the points list as a list where each row is a list of
        # valuesof the same dimension.
        dimensions = zip(*points)
        std = [std_dev(dim) for dim in dimensions]

    norm_points = normalize(points)

    if d == 1 and not is_nested(points):
        sorted_points = sorted(norm_points)
    else:
        # Organize the normalized list as a list where each row is a list of
        # values of the same dimension.
        norm_dimensions = zip(*norm_points)

        # Sort the points by dimension and then return the dimension based
        # list back to the original point form so the sorted list has n
        # records and m dimensions just like the original points list.
        sorted_dimensions = [sorted(dim) for dim in norm_dimensions]
        sorted_points = [list(dim) for dim in zip(*sorted_dimensions)]

    step = n / k
    offset = step / 2
    initial_centroids = sorted_points[offset::step]

    # Perform the clustering then calculate the centroid each point is
    # associated with and the distance between each point and its centroid.
    # Centroids returned here are based on the normalized(see normalize())
    # points.
    centroids, _ = kmeans(norm_points, initial_centroids)
    indexes, distances = compute_clusters(norm_points, centroids)

    # Denormalize the centroids for downstream use. Do this by multiplying
    # each dimension of each centroid by the standard deviation along that
    # dimension that we calculated earlier.
    if d == 1 and not is_nested(points):
        denorm_centroids = [c * std for c in centroids]
    else:
        denorm_centroids = \
            [[dim * s for dim, s in zip(c, std)] for c in centroids]

    return {
        'centroids': denorm_centroids,
        'indexes': indexes,
        'distances': distances,
        'outliers': outliers,
    }


def weighted_counts(points, counts, k):
    """
    Calculate and return the weighted count of each centroid.

    Performs k-means clustering in order to identify the centroids and the
    distances between all points and their centroids. Weights and distances
    for the centroids are then constructed from all the points in that
    centroids cluster. Using those values, the weighted count is computed for
    each of the centroids and returned along with the centroid themselves. The
    points returned are a list of point objects where each point in the list
    has a list of 'values' representing the centroid coordinates and a 'count'
    which is the weighted count as computed for that centroid. Also returned
    is the list of outlier indexes in the 'points' list as determined by the
    k-means algorithm.

    Arguments:
        points - list of n-dimensional points to run k-means on.
        counts - the count for each point in the 'points' list. This list
                 must be of the same length as the 'points' list because the
                 count for a point is determined by accessing both 'points'
                 and 'counts' with the same index.
        k - the number of clusters(k-value) to send to the k-means algorithm.

    Returns:
        centroid_counts - A list of points where each point has 'values' list
        containing the coordinates of the centroid it is associated with and a
        'count' which is an integer representing the weighted count as
        computed for that centroid.
        outliers - A list of indexes of the outliers in 'point's as returned
        by the k-means algorithm.
    """
    result = kmeans_optm(points, k=k)
    outliers = [points[i] for i in result['outliers']]

    dist_weights = defaultdict(lambda: {'dist': [], 'count': []})
    for i, idx in enumerate(result['indexes']):
        dist_weights[idx]['dist'].append(result['distances'][i])
        dist_weights[idx]['count'].append(counts[i])

    centroid_counts = []

    # Determine best count relative to each point in the cluster
    for i, centroid in enumerate(result['centroids']):
        dist_sum = sum(dist_weights[i]['dist'])
        weighted_counts = []
        for j, dist in enumerate(dist_weights[i]['dist']):
            if dist_sum:
                wc = (1 - dist / dist_sum) * dist_weights[i]['count'][j]
            else:
                wc = dist_weights[i]['count'][j]
            weighted_counts.append(wc)

        if is_iterable(centroid):
            values = list(centroid)
        else:
            values = centroid

        centroid_counts.append({
            'values': values,
            'count': int(sum(weighted_counts)),
        })

    return centroid_counts, outliers



import logging
from django.db.models import F
from optparse import make_option
from avocado.management.base import DataFieldCommand

log = logging.getLogger(__name__)

__doc__ = """\
Increments the `data_version` field on DataField instances. This will cause
various cache that depends on this field to be refreshed the next time it is
requested. To pre-cache, use the `avocado cache` command.
"""


class Command(DataFieldCommand):
    help = __doc__

    option_list = DataFieldCommand.option_list + (
        make_option('-i', '--incr', action='store_true', dest='incr_version',
                    default=False, help='Increment `data_version` on '
                    '`DataField` instances'),
    )

    def handle_fields(self, fields, **options):
        "Handles app_label or app_label.model_label formats."

        incr_version = options.get('incr_version')

        if not incr_version:
            print 'Nothing to do.'
            return

        # Increments each field's data version
        updated = fields.update(data_version=F('data_version') + 1)

        print(u'{0} fields have been updated. Cached methods will '
              'lazily refresh their cache the next time they are '
              'accessed.'.format(updated))



import sys
from setuptools import setup, find_packages


install_requires = [
    'django>=1.4,<1.6',
    'modeltree>=1.1.5',
    'South==0.8.1',
    'jsonfield==0.9.4',
]

if sys.version_info < (2, 7):
    install_requires.append('ordereddict>=1.1')


kwargs = {
    # Packages
    'packages': find_packages(exclude=['tests', '*.tests', '*.tests.*',
                                       'tests.*']),
    'include_package_data': True,

    # Dependencies
    'install_requires': install_requires,

    'test_suite': 'test_suite',

    # Test dependencies
    'tests_require': [
        'django-guardian==1.0.4',
        'django-haystack==2.0.0',
        'whoosh==2.4.1',
        'openpyxl>=1.6,<1.7',
        'python-memcached>=1.48',
        'django-objectset>=0.2.2',
    ],

    # Optional dependencies
    'extras_require': {
        # Granular permission
        'permissions': ['django-guardian==1.0.4'],
        # Search
        'search': ['django-haystack==2.0.0'],
        # Includes extra exporter dependencies
        'extras': ['openpyxl>=1.6,<1.7'],
        # Pretty printing of SQL in the admin and for debugging
        'sql': ['sqlparse'],
    },

    # Metadata
    'name': 'avocado',
    'version': __import__('avocado').get_version(),
    'author': 'Byron Ruth',
    'author_email': 'b@devel.io',
    'description': 'Metadata APIs for Django',
    'license': 'BSD',
    'keywords': 'query metadata harvest orm',
    'url': 'http://github.com/cbmi/avocado',
    'classifiers': [
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Framework :: Django',
        'Topic :: Internet :: WWW/HTTP',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Information Technology',
    ],
}

setup(**kwargs)



import sys
import time
import logging
from optparse import make_option
from django.core.management.base import BaseCommand, CommandError
from avocado.models import DataField
from avocado.management.base import DataFieldCommand

log = logging.getLogger(__name__)

# Get all methods that have a `flush` method for clearing the
# internal cache.
CACHED_METHODS = []

for attr_name in dir(DataField):
    attr = getattr(DataField, attr_name)
    # This is crude means of checking for methods that contain a CacheProxy,
    # however that is the only method used below.
    if hasattr(attr, 'flush'):
        CACHED_METHODS.append(attr_name)

CACHED_METHODS = tuple(CACHED_METHODS)


__doc__ = """\
Pre-caches data produced by various DataField methods that are data dependent.
Pass `--flush` to explicitly flush any existing cache for each method.
"""


class Command(DataFieldCommand):
    help = __doc__

    option_list = BaseCommand.option_list + (
        make_option('--flush', action='store_true', help='Flushes existing '
                    'cache for each cached property.'),

        make_option('--method', action='append', dest='methods',
                    default=CACHED_METHODS, help='Select which methods to '
                    'pre-cache. Choices: {0}'.format(
                        ', '.join(CACHED_METHODS))),
    )

    def _progress(self):
        sys.stdout.write('.')
        sys.stdout.flush()

    def handle_fields(self, fields, **options):
        flush = options.get('flush')
        methods = options.get('methods')

        # Validate methods
        for method in methods:
            if method not in CACHED_METHODS:
                raise CommandError('Invalid method {0}. Choices are {1}'
                                   .format(method, ', '.join(CACHED_METHODS)))

        count = 0
        t0 = time.time()

        for f in fields:
            for method in methods:
                func = getattr(f, method)
                if flush:
                    func.flush(f)
                func()
            count += 1
            self._progress()
            log.debug('{0} cache set took {1} seconds'.format(
                f, time.time() - t0))

        print(u'\n{0} fields have been updated ({1} s)'.format(
            count, round(time.time() - t0, 2)))



from .models import Revision


def post_save_revision(instance, created, **kwargs):
    from . import registry
    fields = registry[instance.__class__]['fields']
    Revision.objects.create_revision(instance, fields=fields)


def pre_delete_revision(instance, **kwargs):
    from . import registry
    fields = registry[instance.__class__]['fields']
    Revision.objects.create_revision(instance, fields=fields, deleted=True)
